import ipaddress
import os
import re
import socket
from datetime import datetime
from importlib import metadata

import pymysql
import requests
from dolphie.Modules.Functions import format_number, format_sys_table_memory
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.MetricManager import MetricManager
from dolphie.Modules.MySQL import Database
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Widgets.command_screen import CommandScreen
from dolphie.Widgets.event_log_screen import EventLog
from dolphie.Widgets.modal import CommandModal
from dolphie.Widgets.new_version_modal import NewVersionModal
from dolphie.Widgets.quick_switch import QuickSwitchHostModal
from packaging.version import parse as parse_version
from rich import box
from rich.align import Align
from rich.console import Group
from rich.style import Style
from rich.syntax import Syntax
from rich.table import Table
from sqlparse import format as sqlformat
from textual.app import App

try:
    __package_name__ = metadata.metadata(__package__ or __name__)["Name"]
    __version__ = metadata.version(__package__ or __name__)
except Exception:
    __package_name__ = "Dolphie"
    __version__ = "N/A"


class Dolphie:
    def __init__(self):
        self.app: App = None
        self.app_version = __version__
        self.metric_manager = MetricManager()

        # Config options
        self.user: str = None
        self.password: str = None
        self.host: str = None
        self.port: int = 3306
        self.socket: str = None
        self.ssl: dict = {}
        self.config_file: str = None
        self.host_cache_file: str = None
        self.quick_switch_hosts_file: str = None
        self.debug: bool = False
        self.refresh_interval: int = 1
        self.use_processlist: bool = False
        self.show_idle_threads: bool = False
        self.show_trxs_only: bool = False
        self.show_additional_query_columns: bool = False
        self.sort_by_time_descending: bool = True
        self.hide_dashboard: bool = False
        self.heartbeat_table: str = None
        self.user_filter: str = None
        self.db_filter: str = None
        self.host_filter: str = None
        self.query_time_filter: str = 0
        self.query_filter: str = None

        self.dolphie_start_time: datetime = datetime.now()
        self.worker_start_time: datetime = datetime.now()
        self.worker_previous_start_time: datetime = datetime.now()
        self.worker_job_time: int = 0
        self.processlist_threads: dict = {}
        self.processlist_threads_snapshot: dict = {}
        self.pause_refresh: bool = False
        self.previous_binlog_position: int = 0
        self.previous_replica_sbm: int = 0
        self.quick_switch_hosts: list = []
        self.host_cache: dict = {}
        self.host_cache_from_file: dict = {}
        self.innodb_metrics: dict = {}
        self.global_variables: dict = {}
        self.global_status: dict = {}
        self.binlog_status: dict = {}
        self.replication_status: dict = {}
        self.replication_applier_status: dict = {}
        self.replica_lag_source: str = None
        self.replica_lag: int = None
        self.active_redo_logs: int = None
        self.mysql_host: str = None
        self.quick_switched_connection: bool = False

        # These are for replicas in replication panel
        self.replica_data: dict = {}
        self.replica_connections: dict = {}
        self.replica_tables: dict = {}

        # Panel display states
        self.display_dashboard_panel: bool = False
        self.display_processlist_panel: bool = False
        self.display_replication_panel: bool = False
        self.display_graphs_panel: bool = False

        # Database connection global_variables
        # Main connection is used for Textual's worker thread so it can run asynchronous
        self.main_db_connection: Database = None
        # Secondary connection is for ad-hoc commands that are not a part of the worker thread
        self.secondary_db_connection: Database = None
        self.main_db_connection_id: int = None
        self.secondary_db_connection_id: int = None
        self.use_performance_schema: bool = False
        self.performance_schema_enabled: bool = False
        self.host_is_rds: bool = False
        self.host_is_cluster: bool = False
        self.server_uuid: str = None
        self.mysql_version: str = None
        self.host_distro: str = None

        # Misc
        self.footer_timer = None

    def check_for_update(self):
        # Query PyPI API to get the latest version
        try:
            url = f"https://pypi.org/pypi/{__package_name__}/json"
            response = requests.get(url, timeout=3)

            if response.status_code == 200:
                data = response.json()

                # Extract the latest version from the response
                latest_version = data["info"]["version"]

                # Compare the current version with the latest version
                if parse_version(latest_version) > parse_version(__version__):
                    self.app.push_screen(NewVersionModal(current_version=__version__, latest_version=latest_version))
        except Exception:
            pass

    def is_mysql_version_at_least(self, target):
        parsed_source = parse_version(self.mysql_version)
        parsed_target = parse_version(target)

        return parsed_source >= parsed_target

    def update_footer(self, output, hide=False, temporary=True):
        if len(self.app.screen_stack) > 1:
            return

        footer = self.app.query_one("#footer")

        footer.display = True
        footer.update(output)

        if hide:
            footer.display = False
        elif temporary:
            # Stop existing timer if it exists
            if self.footer_timer and self.footer_timer._active:
                self.footer_timer.stop()
            self.footer_timer = self.app.set_timer(7, lambda: setattr(footer, "display", False))

    def db_connect(self):
        self.main_db_connection = Database(self.host, self.user, self.password, self.socket, self.port, self.ssl)
        self.secondary_db_connection = Database(self.host, self.user, self.password, self.socket, self.port, self.ssl)

        # Reduce any issues with the queries Dolphie runs (mostly targetting only_full_group_by)
        self.main_db_connection.execute("SET SESSION sql_mode = ''")
        self.secondary_db_connection.execute("SET SESSION sql_mode = ''")

        self.mysql_host = self.main_db_connection.fetch_value_from_field("SELECT @@hostname")

        self.main_db_connection_id = self.main_db_connection.fetch_value_from_field("SELECT CONNECTION_ID()")

        query = "SELECT CONNECTION_ID() AS connection_id"
        self.secondary_db_connection_id = self.secondary_db_connection.fetch_value_from_field("SELECT CONNECTION_ID()")

        performance_schema = self.main_db_connection.fetch_value_from_field("SELECT @@performance_schema")
        if performance_schema == 1:
            self.performance_schema_enabled = True

            if not self.use_processlist:
                self.use_performance_schema = True

        version_comment = self.main_db_connection.fetch_value_from_field("SELECT @@version_comment").lower()
        basedir = self.main_db_connection.fetch_value_from_field("SELECT @@basedir")

        aurora_version = None
        query = "SHOW GLOBAL VARIABLES LIKE 'aurora_version'"
        aurora_version_data = self.main_db_connection.fetch_value_from_field(query)
        if aurora_version_data:
            aurora_version = aurora_version_data

        version = self.main_db_connection.fetch_value_from_field("SELECT @@version").lower()
        version_split = version.split(".")

        self.mysql_version = "%s.%s.%s" % (
            version_split[0],
            version_split[1],
            version_split[2].split("-")[0],
        )
        major_version = int(version_split[0])

        # Get proper host version and fork
        if "percona xtradb cluster" in version_comment:
            self.host_distro = "Percona XtraDB Cluster"
            self.host_is_cluster = True
        elif "percona server" in version_comment:
            self.host_distro = "Percona Server"
        elif "mariadb cluster" in version_comment:
            self.host_distro = "MariaDB Cluster"
            self.host_is_cluster = True
        elif "mariadb" in version_comment or "mariadb" in version:
            self.host_distro = "MariaDB"
        elif aurora_version:
            self.host_distro = "Amazon Aurora"
            self.host_is_rds = True
        elif "rdsdb" in basedir:
            self.host_distro = "Amazon RDS"
            self.host_is_rds = True
        else:
            self.host_distro = "MySQL"

        server_uuid_query = "SELECT @@server_uuid"
        if "MariaDB" in self.host_distro and major_version >= 10:
            server_uuid_query = "SELECT @@server_id"
        self.server_uuid = self.main_db_connection.fetch_value_from_field(server_uuid_query)

        # Add host to quick switch hosts file if it doesn't exist
        with open(self.quick_switch_hosts_file, "a+") as file:
            file.seek(0)
            lines = file.readlines()

            if self.port != 3306:
                host = f"{self.host}:{self.port}\n"
            else:
                host = f"{self.host}\n"

            if host not in lines:
                file.write(host)
                self.quick_switch_hosts.append(host[:-1])  # remove the \n

    def command_input_to_variable(self, return_data):
        variable = return_data[0]
        value = return_data[1]
        if value:
            setattr(self, variable, value)

    def toggle_panel(self, panel_name):
        panel = self.app.query_one(f"#panel_{panel_name}")

        new_display = not panel.display
        panel.display = new_display
        setattr(self, f"display_{panel_name}_panel", new_display)
        if panel_name not in ["graphs"]:
            self.app.refresh_panel(panel_name, toggled=True)

    def capture_key(self, key):
        screen_data = None

        if not self.main_db_connection:
            self.update_footer("Database connection must be established before using commands")

            return

        if key == "1":
            self.toggle_panel("dashboard")
        elif key == "2":
            self.toggle_panel("processlist")
            self.app.query_one("#panel_processlist").clear()
        elif key == "3":
            self.toggle_panel("replication")
        elif key == "4":
            self.toggle_panel("graphs")
            self.app.update_graphs("dml")
        elif key == "grave_accent":

            def command_get_input(data):
                host_port = data["host"].split(":")
                self.host = host_port[0]
                self.port = int(host_port[1]) if len(host_port) > 1 else 3306

                password = data.get("password")
                if password:
                    self.password = password

                # Trigger a quick switch connection for the worker thread
                self.quick_switched_connection = True

                self.app.query_one("#main_container").display = False
                self.app.query_one("LoadingIndicator").display = True
                self.app.query_one("#panel_dashboard_queries_qps").display = False

            self.app.push_screen(QuickSwitchHostModal(quick_switch_hosts=self.quick_switch_hosts), command_get_input)

        elif key == "a":
            if self.show_additional_query_columns:
                self.show_additional_query_columns = False
            else:
                self.show_additional_query_columns = True

        elif key == "c":
            self.user_filter = ""
            self.db_filter = ""
            self.host_filter = ""
            self.query_time_filter = ""
            self.query_filter = ""

            self.update_footer("Cleared all filters")

        elif key == "d":
            tables = {}
            all_tables = []

            db_count = self.secondary_db_connection.execute(MySQLQueries.databases)
            databases = self.secondary_db_connection.fetchall()

            # Determine how many tables to provide data
            max_num_tables = 1 if db_count <= 20 else 3

            # Calculate how many databases per table
            row_per_count = db_count // max_num_tables

            # Create dictionary of tables
            for table_counter in range(1, max_num_tables + 1):
                tables[table_counter] = Table(box=box.ROUNDED, show_header=False, style="#52608d")
                tables[table_counter].add_column("")

            # Loop over databases
            db_counter = 1
            table_counter = 1

            # Sort the databases by name
            for database in databases:
                tables[table_counter].add_row(database["SCHEMA_NAME"])
                db_counter += 1

                if db_counter > row_per_count and table_counter < max_num_tables:
                    table_counter += 1
                    db_counter = 1

            # Collect table data into an array
            all_tables = [table_data for table_data in tables.values() if table_data]

            table_grid = Table.grid()
            table_grid.add_row(*all_tables)

            screen_data = Group(
                Align.center("[b]Databases[/b]"),
                Align.center(table_grid),
                Align.center("Total: [b #91abec]%s[/b #91abec]" % db_count),
            )

        elif key == "e":
            if self.is_mysql_version_at_least("8.0"):
                self.app.push_screen(
                    EventLog(self.app_version, f"{self.mysql_host}:{self.port}", self.secondary_db_connection)
                )
            else:
                self.update_footer("Error log command requires MySQL 8")

        elif key == "f":

            def command_get_input(filter_data):
                filter_name = filter_data[0]
                filter_value = filter_data[1]

                if filter_name == "user":
                    self.user_filter = next(
                        (
                            data["user"]
                            for data in self.processlist_threads_snapshot.values()
                            if filter_value == data["user"]
                        ),
                        None,
                    )
                    if not self.user_filter:
                        self.update_footer(
                            f"[indian_red]User[/indian_red] {filter_value}[indian_red] was not found in processlist"
                        )
                        return
                elif filter_name == "database":
                    self.db_filter = next(
                        (
                            data["db"]
                            for data in self.processlist_threads_snapshot.values()
                            if filter_value == data["db"]
                        ),
                        None,
                    )
                    if not self.db_filter:
                        self.update_footer(
                            f"[indian_red]Database[/indian_red] {filter_value}[indian_red] was not found in processlist"
                        )
                        return
                elif filter_name == "host":
                    self.host_filter = next((ip for ip, addr in self.host_cache.items() if filter_value == addr), None)
                    if not self.host_filter:
                        self.update_footer(
                            f"[indian_red]Host[/indian_red] {filter_value}[indian_red] was not found in processlist"
                        )
                        return
                elif filter_name == "query_time":
                    if filter_value.isnumeric():
                        self.query_time_filter = int(filter_value)
                    else:
                        self.update_footer("[bright_red]Query time must be an integer!")
                        return
                elif filter_name == "query_text":
                    self.query_filter = filter_value

                self.update_footer("Now filtering %s by [b #91abec]%s[/b #91abec]" % (filter_name, filter_value))

            self.app.push_screen(
                CommandModal(
                    message="Select which field you'd like to filter by",
                    show_filter_options=True,
                    processlist_data=self.processlist_threads_snapshot,
                ),
                command_get_input,
            )

        elif key == "i":
            if self.show_idle_threads:
                self.show_idle_threads = False
                self.sort_by_time_descending = True

                self.update_footer("Processlist will now hide idle threads")
            else:
                self.show_idle_threads = True
                self.sort_by_time_descending = False

                self.update_footer("Processlist will now show idle threads")

        elif key == "k":

            def command_get_input(thread_id):
                if not thread_id:
                    self.update_footer("[indian_red]You did not specify a Thread ID")
                    return

                if thread_id in self.processlist_threads_snapshot:
                    try:
                        if self.host_is_rds:
                            self.secondary_db_connection.cursor.execute("CALL mysql.rds_kill(%s)" % thread_id)
                        else:
                            self.secondary_db_connection.cursor.execute("KILL %s" % thread_id)

                        self.update_footer("Killed thread [b #91abec]%s[/b #91abec]" % thread_id)
                    except Exception as e:
                        self.update_footer("[b][indian_red]Error killing query[/b]: %s" % e.args[1])
                else:
                    self.update_footer("Thread ID [b #91abec]%s[/b #91abec] does not exist" % thread_id)

            self.app.push_screen(
                CommandModal(message="Specify a Thread ID to kill", processlist_data=self.processlist_threads_snapshot),
                command_get_input,
            )

        elif key == "K":

            def command_get_input(data):
                def execute_kill(thread_id):
                    if self.host_is_rds:
                        self.secondary_db_connection.cursor.execute("CALL mysql.rds_kill(%s)" % thread_id)
                    else:
                        self.secondary_db_connection.cursor.execute("KILL %s" % thread_id)

                kill_type = data[0]
                kill_value = data[1]
                include_sleeping_queries = data[2]

                if not kill_value:
                    self.update_footer("[indian_red]You did not specify a %s" % kill_type)
                    return

                if kill_type == "username":
                    key = "user"
                elif kill_type == "host":
                    key = "host"
                elif kill_type == "time_range":
                    key = "time"
                    if re.search(r"(\d+-\d+)", kill_value):
                        time_range = kill_value.split("-")
                        lower_limit = int(time_range[0])
                        upper_limit = int(time_range[1])

                        if lower_limit > upper_limit:
                            self.update_footer("[indian_red]Invalid time range! Lower limit can't be higher than upper")
                            return
                    else:
                        self.update_footer("[indian_red]Invalid time range")
                        return
                else:
                    self.update_footer("[indian_red]Invalid option")
                    return

                threads_killed = 0
                commands_to_kill = ["Query", "Execute"]
                if include_sleeping_queries:
                    commands_to_kill.append("Sleep")

                for thread_id, thread in self.processlist_threads_snapshot.items():
                    try:
                        if thread["command"] in commands_to_kill:
                            if kill_type == "time_range":
                                if thread["time"] >= lower_limit and thread["time"] <= upper_limit:
                                    execute_kill(thread_id)
                                    threads_killed += 1
                            else:
                                if thread[key] == kill_value:
                                    execute_kill(thread_id)
                                    threads_killed += 1
                    except Exception as e:
                        self.update_footer("[b][indian_red]Error killing query[/b]: %s" % e.args[1])
                        return

                if threads_killed:
                    self.update_footer("Killed [#91abec]%s[/#91abec] threads" % threads_killed)
                else:
                    self.update_footer("No threads were killed")

            self.app.push_screen(
                CommandModal(
                    message="Kill threads based around parameters",
                    show_kill_options=True,
                    processlist_data=self.processlist_threads_snapshot,
                ),
                command_get_input,
            )

        elif key == "l":
            deadlock = ""
            output = re.search(
                r"------------------------\nLATEST\sDETECTED\sDEADLOCK\n------------------------"
                "\n(.*?)------------\nTRANSACTIONS",
                self.secondary_db_connection.fetch_value_from_field(MySQLQueries.innodb_status, "Status"),
                flags=re.S,
            )
            if output:
                deadlock = output.group(1)

                deadlock = deadlock.replace("***", "[#f1fb82]*****[/#f1fb82]")
                screen_data = deadlock
            else:
                screen_data = Align.center("No deadlock detected")

        elif key == "o":
            screen_data = self.secondary_db_connection.fetch_value_from_field(MySQLQueries.innodb_status, "Status")

        elif key == "m":
            table_line_color = "#52608d"

            table_grid = Table.grid()

            table1 = Table(
                box=box.ROUNDED,
                style=table_line_color,
            )

            header_style = Style(bold=True)
            table1.add_column("User", header_style=header_style)
            table1.add_column("Current", header_style=header_style)
            table1.add_column("Total", header_style=header_style)

            self.secondary_db_connection.execute(MySQLQueries.memory_by_user)
            data = self.secondary_db_connection.fetchall()
            for row in data:
                table1.add_row(
                    row["user"],
                    format_sys_table_memory(row["current_allocated"]),
                    format_sys_table_memory(row["total_allocated"]),
                )

            table2 = Table(
                box=box.ROUNDED,
                style=table_line_color,
            )
            table2.add_column("Code Area", header_style=header_style)
            table2.add_column("Current", header_style=header_style)

            self.secondary_db_connection.execute(MySQLQueries.memory_by_code_area)
            data = self.secondary_db_connection.fetchall()
            for row in data:
                table2.add_row(row["code_area"], format_sys_table_memory(row["current_allocated"]))

            table3 = Table(
                box=box.ROUNDED,
                style=table_line_color,
            )
            table3.add_column("Host", header_style=header_style)
            table3.add_column("Current", header_style=header_style)
            table3.add_column("Total", header_style=header_style)

            self.secondary_db_connection.execute(MySQLQueries.memory_by_host)
            data = self.secondary_db_connection.fetchall()
            for row in data:
                table3.add_row(
                    self.get_hostname(row["host"]),
                    format_sys_table_memory(row["current_allocated"]),
                    format_sys_table_memory(row["total_allocated"]),
                )

            table_grid.add_row("", Align.center("[b]Memory Allocation[/b]"), "")
            table_grid.add_row(table1, table3, table2)

            screen_data = Align.center(table_grid)

        elif key == "p":
            if not self.pause_refresh:
                self.pause_refresh = True
                self.update_footer(
                    f"Refresh is paused! Press [b #91abec]{key}[/b #91abec] again to resume",
                    temporary=False,
                )
            else:
                self.pause_refresh = False
                self.update_footer("", hide=True)

        if key == "P":
            if self.use_performance_schema:
                self.use_performance_schema = False
                self.update_footer("Switched to using [b #91abec]Processlist")
            else:
                if self.performance_schema_enabled:
                    self.use_performance_schema = True
                    self.update_footer("Switched to using [b #91abec]Performance Schema")
                else:
                    self.update_footer("[indian_red]You can't switch to Performance Schema because it isn't enabled")

        elif key == "q":
            self.app.exit()

        elif key == "r":

            def command_get_input(refresh_interval):
                if refresh_interval.isnumeric():
                    self.refresh_interval = int(refresh_interval)
                else:
                    self.update_footer("[indian_red]Input must be an integer")

            self.app.push_screen(
                CommandModal(message="Specify refresh interval (in seconds)"),
                command_get_input,
            )

        elif key == "R":
            self.metric_manager = MetricManager()
            active_graph = self.app.query_one("#tabbed_content").active
            self.app.update_graphs(active_graph.split("tab_")[1])
            self.update_footer("Metrics have been reset")

        elif key == "s":
            if self.sort_by_time_descending:
                self.sort_by_time_descending = False
                self.update_footer("Processlist will now sort queries in ascending order")
            else:
                self.sort_by_time_descending = True
                self.update_footer("Processlist will now sort queries in descending order")

        elif key == "t":

            def command_get_input(thread_id):
                if thread_id:
                    if thread_id in self.processlist_threads_snapshot:
                        thread_data = self.processlist_threads_snapshot[thread_id]

                        table = Table(box=box.ROUNDED, show_header=False, style="#52608d")
                        table.add_column("")
                        table.add_column("")

                        table.add_row("[#c5c7d2]Thread ID", str(thread_id))
                        table.add_row("[#c5c7d2]User", thread_data["user"])
                        table.add_row("[#c5c7d2]Host", thread_data["host"])
                        table.add_row("[#c5c7d2]Database", thread_data["db"])
                        table.add_row("[#c5c7d2]Command", thread_data["command"])
                        table.add_row("[#c5c7d2]State", thread_data["state"])
                        table.add_row("[#c5c7d2]Time", thread_data["formatted_time_with_days"])
                        table.add_row("[#c5c7d2]Rows Locked", thread_data["trx_rows_locked"])
                        table.add_row("[#c5c7d2]Rows Modified", thread_data["trx_rows_modified"])

                        # Transaction history
                        transaction_history_title = ""
                        transaction_history_table = Table(box=box.ROUNDED, style="#52608d")
                        if self.is_mysql_version_at_least("5.7") and self.use_performance_schema:
                            query = MySQLQueries.thread_transaction_history.replace(
                                "$placeholder", str(thread_data["mysql_thread_id"])
                            )
                            self.secondary_db_connection.cursor.execute(query)
                            transaction_history = self.secondary_db_connection.fetchall()

                            if transaction_history:
                                transaction_history_title = "[b]Transaction History[/b]"
                                transaction_history_table.add_column("Start Time")
                                transaction_history_table.add_column("Query")

                                for query in transaction_history:
                                    formatted_query = ""
                                    if query["sql_text"]:
                                        formatted_query = Syntax(
                                            re.sub(r"\s+", " ", query["sql_text"]),
                                            "sql",
                                            line_numbers=False,
                                            word_wrap=True,
                                            theme="monokai",
                                            background_color="#000718",
                                        )

                                    transaction_history_table.add_row(
                                        query["start_time"].strftime("%Y-%m-%d %H:%M:%S"), formatted_query
                                    )

                        if (
                            "innodb_thread_concurrency" in self.global_variables
                            and self.global_variables["innodb_thread_concurrency"]
                        ):
                            table.add_row("[#c5c7d2]Tickets", thread_data["trx_concurrency_tickets"])

                        table.add_row("", "")
                        table.add_row("[#c5c7d2]TRX State", thread_data["trx_state"])
                        table.add_row("[#c5c7d2]TRX Operation", thread_data["trx_operation_state"])

                        query = sqlformat(thread_data["query"], reindent_aligned=True)
                        query_db = thread_data["db"]

                        if query:
                            explain_failure = ""
                            explain_data = ""

                            formatted_query = Syntax(
                                query,
                                "sql",
                                line_numbers=False,
                                word_wrap=True,
                                theme="monokai",
                                background_color="#000718",
                            )

                            if query_db:
                                try:
                                    self.secondary_db_connection.cursor.execute("USE %s" % query_db)
                                    self.secondary_db_connection.cursor.execute("EXPLAIN %s" % query)

                                    explain_data = self.secondary_db_connection.fetchall()
                                except pymysql.Error as e:
                                    explain_failure = (
                                        "[b indian_red]EXPLAIN ERROR:[/b indian_red] [indian_red]%s" % e.args[1]
                                    )

                            if explain_data:
                                explain_table = Table(box=box.ROUNDED, style="#52608d")

                                columns = []
                                for row in explain_data:
                                    values = []
                                    for column, value in row.items():
                                        # Exclude possbile_keys field since it takes up too much space
                                        if column == "possible_keys":
                                            continue

                                        # Don't duplicate columns
                                        if column not in columns:
                                            explain_table.add_column(column)
                                            columns.append(column)

                                        if column == "key" and value is None:
                                            value = "[b white on red]NO INDEX[/b white on red]"

                                        if column == "rows":
                                            value = format_number(value)

                                        values.append(str(value))

                                    explain_table.add_row(*values)

                                screen_data = Group(
                                    Align.center(table),
                                    "",
                                    Align.center(formatted_query),
                                    "",
                                    Align.center(explain_table),
                                    "",
                                    Align.center(transaction_history_title),
                                    Align.center(transaction_history_table),
                                )
                            else:
                                screen_data = Group(
                                    Align.center(table),
                                    "",
                                    Align.center(formatted_query),
                                    "",
                                    Align.center(explain_failure),
                                    "",
                                    Align.center(transaction_history_title),
                                    Align.center(transaction_history_table),
                                )
                        else:
                            screen_data = Group(
                                Align.center(table),
                                "",
                                Align.center(transaction_history_title),
                                Align.center(transaction_history_table),
                            )

                        self.app.push_screen(
                            CommandScreen(self.app_version, f"{self.mysql_host}:{self.port}", screen_data)
                        )
                    else:
                        self.update_footer("Thread ID [b #91abec]%s[/b #91abec] does not exist" % thread_id)

            self.app.push_screen(
                CommandModal(
                    message="Specify a Thread ID to display its details",
                    processlist_data=self.processlist_threads_snapshot,
                ),
                command_get_input,
            )

        elif key == "T":
            if self.show_trxs_only:
                self.show_trxs_only = False
                self.update_footer("Processlist will now no longer only show threads that have an active transaction")
            else:
                self.show_trxs_only = True
                self.update_footer("Processlist will now only show threads that have an active transaction")

        elif key == "u":
            user_stat_data = self.create_user_stats_table()
            if user_stat_data:
                screen_data = Align.center(user_stat_data)
            else:
                self.update_footer(
                    "[b indian_red]Cannot use this command![/b indian_red] It requires Performance Schema to be enabled"
                )

        elif key == "v":

            def command_get_input(input_variable):
                table_grid = Table.grid()
                table_counter = 1
                variable_counter = 1
                row_counter = 1
                variable_num = 1
                all_tables = []
                tables = {}
                display_global_variables = {}

                variable_data = self.secondary_db_connection.fetch_data("variables")
                for variable, value in variable_data.items():
                    if input_variable:
                        if input_variable in variable:
                            display_global_variables[variable] = variable_data[variable]
                    else:
                        display_global_variables[variable] = variable_data[variable]

                max_num_tables = 1 if len(display_global_variables) <= 50 else 2

                # Create the number of tables we want
                while table_counter <= max_num_tables:
                    tables[table_counter] = Table(box=box.ROUNDED, show_header=False, style="#52608d")
                    tables[table_counter].add_column("")
                    tables[table_counter].add_column("")

                    table_counter += 1

                # Calculate how many global_variables per table
                row_per_count = len(display_global_variables) // max_num_tables

                # Loop global_variables
                for variable, value in display_global_variables.items():
                    tables[variable_num].add_row("[#c5c7d2]%s" % variable, str(value))

                    if variable_counter == row_per_count and row_counter != max_num_tables:
                        row_counter += 1
                        variable_counter = 0
                        variable_num += 1

                    variable_counter += 1

                # Put all the variable data from dict into an array
                all_tables = [table_data for table_data in tables.values() if table_data]

                # Add the data into a single tuple for add_row
                if display_global_variables:
                    table_grid.add_row(*all_tables)
                    screen_data = Align.center(table_grid)

                    self.app.push_screen(CommandScreen(self.app_version, f"{self.mysql_host}:{self.port}", screen_data))
                else:
                    if input_variable:
                        self.update_footer("No variable(s) found that match [b #91abec]%s[/b #91abec]" % input_variable)

            self.app.push_screen(
                CommandModal(message="Specify a variable to wildcard search\n[dim](leave blank for all)[/dim]"),
                command_get_input,
            )

        elif key == "z":
            if self.host_cache:
                table = Table(box=box.ROUNDED, style="#52608d")
                table.add_column("Host/IP")
                table.add_column("Hostname (if resolved)")

                for ip, addr in self.host_cache.items():
                    if ip:
                        table.add_row(ip, addr)

                screen_data = Group(
                    Align.center("[b]Host Cache[/b]"),
                    Align.center(table),
                    Align.center("Total: [b #91abec]%s" % len(self.host_cache)),
                )
            else:
                screen_data = Align.center("\nThere are currently no hosts resolved")

        elif key == "question_mark":
            table_line_color = "#52608d"

            keys = {
                "`": "Quickly connect to another host",
                "a": "Toggle additional processlist columns",
                "c": "Clear all filters set",
                "d": "Display all databases",
                "e": "Display error log from Performance Schema",
                "f": "Filter processlist by a supported option",
                "i": "Toggle displaying idle threads",
                "k": "Kill a thread by its ID",
                "K": "Kill a thread by a supported option",
                "l": "Display the most recent deadlock",
                "o": "Display output from SHOW ENGINE INNODB STATUS",
                "m": "Display memory usage",
                "p": "Pause refreshing",
                "P": "Switch between using SHOW PROCESSLIST/Performance Schema for processlist panel",
                "q": "Quit",
                "r": "Set the refresh interval",
                "R": "Reset all metrics",
                "t": "Display details of a thread along with an EXPLAIN of its query",
                "T": "Toggle displaying threads that only have an active transaction",
                "s": "Sort processlist by time in descending/ascending order",
                "u": "List active connected users and their statistics",
                "v": "Variable wildcard search sourced from SHOW GLOBAL VARIABLES",
                "z": "Display all entries in the host cache",
            }

            table_keys = Table(box=box.HORIZONTALS, style=table_line_color, title="Commands", title_style="bold")
            table_keys.add_column("Key", justify="center", style="b #91abec")
            table_keys.add_column("Description")

            for key, description in keys.items():
                table_keys.add_row(key, description)

            panels = {
                "1": "Show/hide Dashboard",
                "2": "Show/hide Processlist",
                "3": "Show/hide Replication/Replicas",
                "4": "Show/hide Graph Metrics",
            }
            table_panels = Table(box=box.HORIZONTALS, style=table_line_color, title="Panels", title_style="bold")
            table_panels.add_column("Key", justify="center", style="b #91abec")
            table_panels.add_column("Description")
            for key, description in sorted(panels.items()):
                table_panels.add_row(key, description)

            datapoints = {
                "Read Only": "If the host is in read-only mode",
                "Read Hit": "The percentage of how many reads are from InnoDB buffer pool compared to from disk",
                "Lag": (
                    "Retrieves metric from: Default -> SHOW SLAVE STATUS, HB -> Heartbeat table, PS -> Performance"
                    " Schema"
                ),
                "Chkpt Age": (
                    "This depicts how close InnoDB is before it starts to furiously flush dirty data to disk "
                    "(Lower is better)"
                ),
                "AHI Hit": (
                    "The percentage of how many lookups there are from Adapative Hash Index compared to it not"
                    " being used"
                ),
                "Diff": "This is the size difference of the binary log between each refresh interval",
                "Cache Hit": "The percentage of how many binary log lookups are from cache instead of from disk",
                "History List": "History list length (number of un-purged row changes in InnoDB's undo logs)",
                "QPS": "Queries per second from Com_queries in SHOW GLOBAL STATUS",
                "Latency": "How much time it takes to receive data from the host for each refresh interval",
                "Threads": "Con = Connected, Run = Running, Cac = Cached from SHOW GLOBAL STATUS",
                "Speed": "How many seconds were taken off of replication lag from the last refresh interval",
                "Tickets": "Relates to innodb_concurrency_tickets variable",
            }

            table_terminology = Table(
                box=box.HORIZONTALS, style=table_line_color, title="Terminology", title_style="bold"
            )
            table_terminology.add_column("Datapoint", style="#91abec")
            table_terminology.add_column("Description")
            for datapoint, description in sorted(datapoints.items()):
                table_terminology.add_row(datapoint, description)

            screen_data = Group(
                Align.center(table_keys),
                "",
                Align.center(table_panels),
                "",
                Align.center(table_terminology),
                "",
                Align.center(
                    "[#bbc8e8][b]Note[/b]: Textual puts your terminal in application mode which disables selecting"
                    " text.\nTo see how to select text on your terminal, visit: https://tinyurl.com/dolphie-select-text"
                ),
            )

        if screen_data:
            self.app.push_screen(CommandScreen(self.app_version, f"{self.mysql_host}:{self.port}", screen_data))

    def create_user_stats_table(self):
        table = Table(header_style="bold white", box=box.ROUNDED, style="#52608d")

        columns = {}
        user_stats = {}

        if self.performance_schema_enabled:
            self.secondary_db_connection.execute(MySQLQueries.ps_user_statisitics)

            columns.update(
                {
                    "User": {"field": "user", "format_number": False},
                    "Active": {"field": "current_connections", "format_number": True},
                    "Total": {"field": "total_connections", "format_number": True},
                    "Rows Read": {"field": "rows_read", "format_number": True},
                    "Rows Sent": {"field": "rows_sent", "format_number": True},
                    "Rows Updated": {"field": "rows_affected", "format_number": True},
                    "Tmp Tables": {"field": "created_tmp_tables", "format_number": True},
                    "Tmp Disk Tables": {"field": "created_tmp_disk_tables", "format_number": True},
                    "Plugin": {"field": "plugin", "format_number": False},
                    "Password Expire": {"field": "password_expires_in", "format_number": False},
                }
            )
        else:
            return False

        users = self.secondary_db_connection.fetchall()
        for user in users:
            username = user["user"]
            user_stats[username] = {
                "user": username,
                "total_connections": user["total_connections"],
                "current_connections": user["current_connections"],
                "password_expires_in": user["password_expires_in"],
                "plugin": user["plugin"],
                "rows_affected": user["sum_rows_affected"],
                "rows_sent": user["sum_rows_sent"],
                "rows_read": user["sum_rows_examined"],
                "created_tmp_disk_tables": user["sum_created_tmp_disk_tables"],
                "created_tmp_tables": user["sum_created_tmp_tables"],
            }

        for column, data in columns.items():
            table.add_column(column, no_wrap=True)

        for user_data in user_stats.values():
            row_values = []
            for column, data in columns.items():
                value = user_data.get(data["field"])

                if data["format_number"]:
                    row_values.append(format_number(value) if value else "")
                else:
                    row_values.append(value or "")

            table.add_row(*row_values)

        return table if user_stats else False

    def load_host_cache_file(self):
        if os.path.exists(self.host_cache_file):
            with open(self.host_cache_file) as file:
                for line in file:
                    line = line.strip()
                    error_message = f"Host cache entry '{line}' is not properly formatted! Format: ip=hostname"

                    if "=" not in line:
                        raise ManualException(error_message)

                    host, hostname = line.split("=", maxsplit=1)
                    host = host.strip()
                    hostname = hostname.strip()

                    if not host or not hostname:
                        raise ManualException(error_message)

                    self.host_cache_from_file[host] = hostname

    def get_hostname(self, host):
        if host in self.host_cache:
            return self.host_cache[host]

        if self.host_cache_from_file and host in self.host_cache_from_file:
            self.host_cache[host] = self.host_cache_from_file[host]
            return self.host_cache_from_file[host]

        try:
            ipaddress.IPv4Network(host)
            hostname = socket.gethostbyaddr(host)[0]
            self.host_cache[host] = hostname
        except (ValueError, socket.error):
            self.host_cache[host] = host
            hostname = host

        return hostname

    def massage_metrics_data(self):
        if self.is_mysql_version_at_least("8.0"):
            # If we're using MySQL 8, we need to fetch the checkpoint age from the performance schema if it's not
            # available in global status
            if not self.global_status.get("Innodb_checkpoint_age"):
                self.global_status["Innodb_checkpoint_age"] = self.main_db_connection.fetch_value_from_field(
                    MySQLQueries.checkpoint_age, "checkpoint_age"
                )

            if self.is_mysql_version_at_least("8.0.30"):
                active_redo_logs_count = self.main_db_connection.fetch_value_from_field(
                    MySQLQueries.active_redo_logs, "count"
                )
                self.global_status["Active_redo_log_count"] = active_redo_logs_count

        # If the server doesn't support Innodb_lsn_current, use Innodb_os_log_written instead
        # which has less precision, but it's good enough
        if not self.global_status.get("Innodb_lsn_current"):
            self.global_status["Innodb_lsn_current"] = self.global_status["Innodb_os_log_written"]

    def fetch_replication_data(self, replica_cursor=None):
        if self.heartbeat_table:
            query = MySQLQueries.heartbeat_replica_lag
            replica_lag_source = "HB"
        elif self.is_mysql_version_at_least("8.0") and self.performance_schema_enabled:
            query = MySQLQueries.ps_replica_lag
            replica_lag_source = "PS"
        else:
            query = MySQLQueries.replication_status
            replica_lag_source = None

        if replica_cursor:
            replica_cursor.execute(query)
            replica_lag_data = replica_cursor.fetchone()
        else:
            # Determine if this server is a replica or not
            self.main_db_connection.execute(MySQLQueries.replication_status)
            replica_lag_data = self.main_db_connection.fetchone()
            self.replication_status = replica_lag_data

            if self.replication_status:
                # Use a better way to detect replication lag if available
                if replica_lag_source:
                    self.main_db_connection.execute(query)
                    replica_lag_data = self.main_db_connection.fetchone()

                # If we're using MySQL 8, fetch the replication applier status data
                self.replication_applier_status = None
                if (
                    self.is_mysql_version_at_least("8.0")
                    and self.display_replication_panel
                    and self.global_variables.get("slave_parallel_workers", 0) > 1
                ):
                    self.main_db_connection.execute(MySQLQueries.replication_applier_status)
                    self.replication_applier_status = self.main_db_connection.fetchall()

        if replica_lag_data:
            replica_lag = int(replica_lag_data["Seconds_Behind_Master"])

            if replica_lag < 0:
                replica_lag = 0
        else:
            replica_lag = 0

        if replica_cursor:
            return replica_lag_source, replica_lag
        else:
            # Save the previous replica lag for to determine Speed data point
            self.previous_replica_sbm = self.replica_lag

            self.replica_lag_source = replica_lag_source
            self.replica_lag = replica_lag



from dataclasses import dataclass


@dataclass
class MySQLQueries:
    pl_query: str = """
        SELECT
            id,
            IFNULL(User, "")                    AS user,
            IFNULL(Host, "")                    AS host,
            IFNULL(db, "")                      AS db,
            IFNULL(Command, "")                 As command,
            IFNULL(Time, "0")                   AS time,
            IFNULL(Info, "")                    AS query,
            IFNULL(State, "")                   AS state,
            IFNULL(trx_query, "")               AS trx_query,
            IFNULL(trx_state, "")               AS trx_state,
            IFNULL(trx_operation_state, "")     AS trx_operation_state,
            IFNULL(trx_rows_locked, "0")        AS trx_rows_locked,
            IFNULL(trx_rows_modified, "0")      AS trx_rows_modified,
            IFNULL(trx_concurrency_tickets, "") AS trx_concurrency_tickets
        FROM
            information_schema.PROCESSLIST pl
            LEFT JOIN information_schema.innodb_trx ON trx_mysql_thread_id = pl.Id
        WHERE 1 $placeholder
    """

    ps_query: str = """
        SELECT
            processlist_id                      AS id,
            IFNULL(thread_id, "0")              AS mysql_thread_id,
            IFNULL(processlist_user, "")        AS user,
            IFNULL(processlist_host, "")        AS host,
            IFNULL(processlist_db, "")          AS db,
            IFNULL(processlist_command, "")     As command,
            IFNULL(processlist_time, "0")       AS time,
            IFNULL(processlist_info, "")        AS query,
            IFNULL(processlist_state, "")       AS state,
            IFNULL(trx_query, "")               AS trx_query,
            IFNULL(trx_state, "")               AS trx_state,
            IFNULL(trx_operation_state, "")     AS trx_operation_state,
            IFNULL(trx_rows_locked, "0")        AS trx_rows_locked,
            IFNULL(trx_rows_modified, "0")      AS trx_rows_modified,
            IFNULL(trx_concurrency_tickets, "") AS trx_concurrency_tickets
        FROM
            performance_schema.threads t
            LEFT JOIN information_schema.innodb_trx tx ON trx_mysql_thread_id = t.processlist_id
        WHERE
            processlist_id IS NOT NULL AND
            processlist_time IS NOT NULL AND
            processlist_command != 'Daemon'
            $placeholder
    """
    ps_replica_lag: str = """
        SELECT
            IFNULL(TIMESTAMPDIFF(
                SECOND,
                MIN(APPLYING_TRANSACTION_ORIGINAL_COMMIT_TIMESTAMP),
            NOW()), "0") AS Seconds_Behind_Master
        FROM
            performance_schema.replication_applier_status_by_worker
        WHERE
            APPLYING_TRANSACTION != ''
    """
    heartbeat_replica_lag: str = """
        SELECT
            TIMESTAMPDIFF(SECOND, MAX(ts), NOW()) AS Seconds_Behind_Master
        FROM
            $placeholder
    """
    ps_find_replicas: str = """
        SELECT
            processlist_id   AS id,
            processlist_user AS user,
            processlist_host AS host
        FROM
            performance_schema.threads
        WHERE
            processlist_command LIKE 'Binlog Dump%'
    """
    pl_find_replicas: str = """
        SELECT
            Id   AS id,
            User AS user,
            Host AS host
        FROM
            information_schema.PROCESSLIST
        WHERE
            Command Like 'Binlog Dump%'
    """
    ps_user_statisitics: str = """
        SELECT
            u.user AS user,
            total_connections,
            current_connections,
            CONVERT(SUM(sum_rows_affected), UNSIGNED) AS sum_rows_affected,
            CONVERT(SUM(sum_rows_sent), UNSIGNED) AS sum_rows_sent,
            CONVERT(SUM(sum_rows_examined), UNSIGNED) AS sum_rows_examined,
            CONVERT(SUM(sum_created_tmp_disk_tables), UNSIGNED) AS sum_created_tmp_disk_tables,
            CONVERT(SUM(sum_created_tmp_tables), UNSIGNED) AS sum_created_tmp_tables,
            plugin,
            CASE
                WHEN (password_lifetime IS NULL OR password_lifetime = 0) AND @@default_password_lifetime = 0 THEN "N/A"
                ELSE CONCAT(
                    CAST(IFNULL(password_lifetime, @@default_password_lifetime) as signed) +
                    CAST(DATEDIFF(password_last_changed, NOW()) as signed),
                    " days"
                )
            END AS password_expires_in
        FROM
            performance_schema.users u
            JOIN performance_schema.events_statements_summary_by_user_by_event_name ess ON u.user = ess.user
            JOIN mysql.user mysql_user ON mysql_user.user = u.user
        WHERE
            current_connections != 0
        GROUP BY
            user
        ORDER BY
            current_connections DESC
    """
    error_log: str = """
        SELECT
            logged AS timestamp,
            prio AS level,
            subsystem,
            data AS message
        FROM
            performance_schema.error_log
        WHERE
            data != 'Could not open log file.'
            $placeholder
        ORDER BY timestamp
    """
    memory_by_user: str = """
        SELECT
            user,
            current_allocated,
            total_allocated
        FROM
            sys.memory_by_user_by_current_bytes
        WHERE
            user != "background"
    """
    memory_by_code_area: str = """
        SELECT
            SUBSTRING_INDEX( event_name, '/', 2 ) AS code_area,
            sys.format_bytes (
            SUM( current_alloc )) AS current_allocated
        FROM
            sys.x$memory_global_by_current_bytes
        GROUP BY
            SUBSTRING_INDEX( event_name, '/', 2 )
        ORDER BY
            SUM( current_alloc ) DESC
    """
    memory_by_host: str = """
        SELECT
            host,
            current_allocated,
            total_allocated
        FROM
            sys.memory_by_host_by_current_bytes
        WHERE
            host != "background"
    """
    databases: str = """
        SELECT
            SCHEMA_NAME
        FROM
            information_schema.SCHEMATA
        ORDER BY
            SCHEMA_NAME
    """
    innodb_metrics: str = """
        SELECT
            NAME,
            COUNT
        FROM
            information_schema.INNODB_METRICS
    """
    checkpoint_age: str = """
        SELECT
            STORAGE_ENGINES ->> '$."InnoDB"."LSN"' - STORAGE_ENGINES ->> '$."InnoDB"."LSN_checkpoint"' AS checkpoint_age
        FROM
            performance_schema.log_status
    """
    active_redo_logs: str = """
        SELECT
            COUNT(*) AS count
        FROM
            performance_schema.file_instances
        WHERE
            file_name LIKE '%innodb_redo/%' AND
            file_name NOT LIKE '%_tmp'
    """
    thread_transaction_history: str = """
        SELECT
            DATE_SUB(
                NOW(),
                INTERVAL (
                    SELECT variable_value
                    FROM performance_schema.global_status
                    WHERE variable_name = 'UPTIME'
                ) - TIMER_START * 10e-13 SECOND
            ) AS start_time,
            sql_text
        FROM
            performance_schema.events_statements_history
        WHERE
            nesting_event_id = (
                SELECT EVENT_ID
                FROM performance_schema.events_transactions_current t
                WHERE t.thread_id = $placeholder
            )
        ORDER BY
            event_id;
    """
    replication_applier_status: str = """
        SELECT
            worker_id,
            FORMAT_PICO_TIME(
                (applier_status.LAST_APPLIED_TRANSACTION_END_APPLY_TIMESTAMP -
                applier_status.LAST_APPLIED_TRANSACTION_START_APPLY_TIMESTAMP) * 1000000000000
            ) AS apply_time,
            applier_status.LAST_APPLIED_TRANSACTION AS last_applied_transaction,
            CONVERT(SUM(thread_events.COUNT_STAR), UNSIGNED) AS total_thread_events
        FROM
            `performance_schema`.replication_applier_status_by_worker applier_status
        JOIN
            `performance_schema`.events_transactions_summary_by_thread_by_event_name thread_events ON
            applier_status.THREAD_ID = thread_events.THREAD_ID
        WHERE
            applier_status.THREAD_ID IN (
                SELECT THREAD_ID FROM `performance_schema`.replication_applier_status_by_worker
            )
        GROUP BY worker_id
        WITH ROLLUP
        ORDER BY worker_id
    """
    status: str = "SHOW GLOBAL STATUS"
    variables: str = "SHOW GLOBAL VARIABLES"
    binlog_status: str = "SHOW MASTER STATUS"
    replication_status: str = "SHOW SLAVE STATUS"
    innodb_status: str = "SHOW ENGINE INNODB STATUS"



from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Input, Label
from textual_autocomplete import AutoComplete, Dropdown, DropdownItem


class QuickSwitchHostModal(ModalScreen):
    CSS = """
        QuickSwitchHostModal > Vertical {
            background: #121626;
            border: thick #212941;
            height: auto;
            width: 70;
        }
        QuickSwitchHostModal > Vertical > * {
            width: auto;
            height: auto;
            align: center middle;
        }
        QuickSwitchHostModal Label {
            text-style: bold;
            width: 100%;
            content-align: center middle;
            padding-bottom: 1;
        }
        QuickSwitchHostModal Input {
            width: 100% !important;
            content-align: center middle;
        }
        QuickSwitchHostModal .main_container {
            width: 100%;
            content-align: center middle;
        }
        QuickSwitchHostModal AutoComplete {
            width: 100%;
            height: auto;
        }
    """
    BINDINGS = [
        Binding("escape", "app.pop_screen", "", show=False),
    ]

    def __init__(self, quick_switch_hosts):
        super().__init__()

        self.dropdown_items = []
        if quick_switch_hosts:
            self.dropdown_items = [DropdownItem(id) for id in sorted(quick_switch_hosts)]

    def compose(self) -> ComposeResult:
        with Vertical():
            with Vertical(classes="main_container"):
                yield Label("Quick Switch Host")
                yield AutoComplete(
                    Input(id="host", placeholder="Start typing to search for a host"),
                    Dropdown(id="dropdown_items", items=self.dropdown_items),
                )
                yield Input(id="password", placeholder="Password (empty for current)", password=True)
            with Horizontal(classes="button_container"):
                yield Button("Submit", id="submit", variant="primary")
                yield Button("Cancel", id="cancel")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "submit":
            host = self.query_one("#host", Input)
            password = self.query_one("#password", Input)

            if host.value:
                self.dismiss({"host": host.value, "password": password.value})
        else:
            self.app.pop_screen()

    @on(Input.Submitted, "#password")
    def on_input_submitted(self):
        self.query_one("#submit", Button).press()



from dolphie.Modules.Queries import MySQLQueries
from dolphie.Widgets.topbar import TopBar
from textual import events, on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.screen import Screen
from textual.widgets import DataTable, Input, Label, Switch


class EventLog(Screen):
    CSS = """
        EventLog Horizontal {
            height: auto;
            padding-top: 1;
            align: center top;
            background: #000718;
            width: 100%;
        }
        EventLog Horizontal > Label {
            color: #bbc8e8;
            text-style: bold;
            margin-right: -1;
        }
        #info {
            padding-top: 1;
            width: 100%;
            text-align: center;
            text-style: bold;
        }
        #search {
            background: #030918;
            content-align: right middle;
            padding-left: 1;
            margin: 0;
            border: none;
        }
    """

    def __init__(self, app_version, host, db):
        super().__init__()

        self.app_version = app_version
        self.host = host
        self.db = db

        self.levels = {
            "system": {"active": True, "sql": "prio = 'System'"},
            "warning": {"active": True, "sql": "prio = 'Warning'"},
            "error": {"active": True, "sql": "prio = 'Error'"},
        }

    def on_mount(self):
        datatable = self.query_one(DataTable)
        datatable.focus()
        datatable.styles.overflow_x = "auto"

        self.update_datatable()

    @on(events.Key)
    def on_keypress(self, event: events.Key):
        if event.key == "q":
            if self.screen.is_attached:
                self.app.pop_screen()
        elif event.key == "1":
            table = self.query_one(DataTable)
            table.move_cursor(row=0)
        elif event.key == "2":
            table = self.query_one(DataTable)
            table.move_cursor(row=table.row_count - 1)

    def compose(self) -> ComposeResult:
        yield TopBar(
            app_version=self.app_version,
            host=self.host,
            help="Press [b]q[/b] to return, [b]1[/b]=top/[b]2[/b]=bottom of events",
        )

        with Horizontal():
            switch_options = [("System", "system"), ("Warning", "warning"), ("Error", "error")]
            for label, switch_id in switch_options:
                yield Label(label)
                yield Switch(animate=False, id=switch_id, value=True)

        yield Label("", id="info")
        yield Input(id="search", placeholder="Search (hit enter when ready)")
        with Container():
            yield DataTable(show_cursor=False)

    @on(Input.Submitted, "#search")
    def event_search(self):
        self.update_datatable()

    @on(Switch.Changed)
    def switch_changed(self, event: Switch.Changed):
        self.levels[event.switch.id]["active"] = event.value

        self.update_datatable()

    def update_datatable(self):
        active_sql_list = [data["sql"] for data in self.levels.values() if data["active"]]
        where_clause = " OR ".join(active_sql_list)

        search_text = self.query_one("#search", Input)
        if search_text.value:
            where_clause = f"({where_clause}) AND (data LIKE '%{search_text.value}%')"

        table = self.query_one(DataTable)
        info = self.query_one("#info", Label)

        table.clear(columns=True)
        table.add_column("Time")
        table.add_column("Level")

        if where_clause:
            query = MySQLQueries.error_log.replace("$placeholder", f"AND ({where_clause})")
            event_count = self.db.execute(query)
            data = self.db.fetchall()

            if data:
                table.add_column(f"Event ({event_count})")

                info.display = False
                table.display = True
                search_text.display = True

                for row in data:
                    level_color = ""
                    if row["level"] == "Error":
                        level_color = "white on red"
                    elif row["level"] == "Warning":
                        level_color = "#f1fb82"

                    level = row["level"]
                    if level_color:
                        level = f"[{level_color}]{row['level']}[/{level_color}]"

                    timestamp = f"[#858A97]{row['timestamp'].strftime('%Y-%m-%d %H:%M:%S')}[/#858A97]"

                    table.add_row(timestamp, level, row["message"])
            else:
                table.display = False
                search_text.display = False
                info.display = True
                info.update("No events found")
        else:
            table.display = False
            search_text.display = False
            info.display = True
            info.update("Toggle the switches above to filter what events you'd like to see")



import re
from datetime import timedelta

import pymysql
from dolphie import Dolphie
from dolphie.Modules.Queries import MySQLQueries
from rich import box
from rich.align import Align
from rich.console import Group
from rich.panel import Panel
from rich.style import Style
from rich.table import Table


def create_panel(dolphie: Dolphie) -> Panel:
    if (
        dolphie.display_replication_panel
        and not dolphie.replica_data
        and not dolphie.replication_status
        and not dolphie.host_is_cluster
    ):
        return "[#f1fb82]No data to display![/#f1fb82] This host is not a replica and has no replicas connected"

    def create_replica_panel():
        if not dolphie.replica_tables and not dolphie.replica_data:
            return None

        if dolphie.replica_tables:
            table_grid = Table.grid()

            num_replicas = len(dolphie.replica_tables)
            for i in range(0, num_replicas, 2):
                table_grid.add_row(*[table for _, table in sorted(list(dolphie.replica_tables.items()))[i : i + 2]])

            content = table_grid
        elif dolphie.replica_data:
            num_replicas = len(dolphie.replica_data)
            content = "\nLoading...\n"

        title = "Replica" if num_replicas == 1 else "Replicas"
        return Panel(
            Align.center(content),
            title=f"[b #e9e9e9]{num_replicas} {title}",
            box=box.HORIZONTALS,
            border_style="#6171a6",
        )

    def create_cluster_panel():
        if not dolphie.host_is_cluster:
            return None

        return Panel(
            Align.center(
                f"\n[b #bbc8e8]State[/b #bbc8e8]  {dolphie.global_status.get('wsrep_local_state_comment', 'N/A')}\n"
            ),
            title="[b #e9e9e9]Cluster",
            box=box.HORIZONTALS,
            border_style="#6171a6",
        )

    def create_replication_panel():
        if not dolphie.replication_status:
            return None

        replication_variables = ""
        available_replication_variables = {
            "binlog_transaction_dependency_tracking": "dependency_tracking",
            "slave_parallel_type": "parallel_type",
            "slave_parallel_workers": "parallel_workers",
            "slave_preserve_commit_order": "preserve_commit_order",
        }

        for setting_variable, setting_display_name in available_replication_variables.items():
            value = dolphie.global_variables.get(setting_variable, "N/A")
            replication_variables += f"[b #bbc8e8]{setting_display_name}[/b #bbc8e8] {value}  "
        replication_variables = replication_variables.strip()

        table_thread_applier_status = Table()
        if dolphie.replication_applier_status:
            table_thread_applier_status = Table(title_style=Style(bold=True), style="#6171a6", box=box.ROUNDED)
            table_thread_applier_status.add_column("Worker", justify="center")
            table_thread_applier_status.add_column("Usage", min_width=6)
            table_thread_applier_status.add_column("Apply Time")
            table_thread_applier_status.add_column("Last Applied Transaction")

            for row in dolphie.replication_applier_status:
                # We use ROLLUP in the query, so the first row is the total for thread_events
                if not row["worker_id"]:
                    total_thread_events = row["total_thread_events"]
                    continue

                last_applied_transaction = ""
                if row["last_applied_transaction"]:
                    source_id_split = row["last_applied_transaction"].split("-")[4].split(":")[0]
                    transaction_id = row["last_applied_transaction"].split(":")[1]
                    last_applied_transaction = f"…[#969aad]{source_id_split}[/#969aad]:{transaction_id}"

                table_thread_applier_status.add_row(
                    str(row["worker_id"]),
                    str(round(100 * (row["total_thread_events"] / total_thread_events), 2)) + "%",
                    row["apply_time"],
                    last_applied_transaction,
                )

        table_grid_replication = Table.grid()
        table_grid_replication.add_row(create_table(dolphie), table_thread_applier_status)

        return Panel(
            Group(Align.center(replication_variables), Align.center(table_grid_replication)),
            title="[b #e9e9e9]Replication",
            box=box.HORIZONTALS,
            border_style="#6171a6",
        )

    group_panels = [
        create_cluster_panel(),
        create_replication_panel(),
        create_replica_panel(),
    ]

    return Group(*[panel for panel in group_panels if panel])


def create_table(dolphie: Dolphie, data=None, dashboard_table=False, replica_thread_id=None) -> Table:
    table_title_style = Style(bold=True)
    table_line_color = "#6171a6"
    table_box = box.ROUNDED

    # When replica_thread_id is specified, that means we're creating a table for a replica and not replication
    if replica_thread_id:
        if dolphie.replica_connections[replica_thread_id]["previous_sbm"] is not None:
            replica_previous_replica_sbm = dolphie.replica_connections[replica_thread_id]["previous_sbm"]

        db_cursor = dolphie.replica_connections[replica_thread_id]["cursor"]
        sbm_source, data["Seconds_Behind_Master"] = dolphie.fetch_replication_data(replica_cursor=db_cursor)
    else:
        data = dolphie.replication_status
        replica_previous_replica_sbm = dolphie.previous_replica_sbm

        sbm_source = dolphie.replica_lag_source
        data["Seconds_Behind_Master"] = dolphie.replica_lag

    if data["Slave_IO_Running"].lower() == "no":
        data["Slave_IO_Running"] = "[#fc7979]NO[/#fc7979]"
    else:
        data["Slave_IO_Running"] = "[#54efae]Yes[/#54efae]"

    if data["Slave_SQL_Running"].lower() == "no":
        data["Slave_SQL_Running"] = "[#fc7979]NO[/#fc7979]"
    else:
        data["Slave_SQL_Running"] = "[#54efae]Yes[/#54efae]"

    speed = 0
    if data["Seconds_Behind_Master"] is not None:
        replica_sbm = data["Seconds_Behind_Master"]

        if replica_thread_id:
            dolphie.replica_connections[replica_thread_id]["previous_sbm"] = replica_sbm

        if replica_previous_replica_sbm and replica_sbm < replica_previous_replica_sbm:
            speed = round((replica_previous_replica_sbm - replica_sbm) / dolphie.worker_job_time)

        if replica_sbm >= 20:
            lag = "[#fc7979]%s" % "{:0>8}[/#fc7979]".format(str(timedelta(seconds=replica_sbm)))
        elif replica_sbm >= 10:
            lag = "[#f1fb82]%s[/#f1fb82]" % "{:0>8}".format(str(timedelta(seconds=replica_sbm)))
        else:
            lag = "[#54efae]%s[/#54efae]" % "{:0>8}".format(str(timedelta(seconds=replica_sbm)))

    data["Master_Host"] = dolphie.get_hostname(data["Master_Host"])
    mysql_gtid_enabled = False
    mariadb_gtid_enabled = False
    gtid_status = "OFF"
    if "Executed_Gtid_Set" in data and data["Executed_Gtid_Set"]:
        mysql_gtid_enabled = True
        gtid_status = "ON"
    if "Using_Gtid" in data and data["Using_Gtid"] != "No":
        mariadb_gtid_enabled = True
        gtid_status = data["Using_Gtid"]

    table_title = ""
    if dashboard_table:
        table_title = "Replication"

    table = Table(
        show_header=False,
        box=table_box,
        title=table_title,
        title_style=table_title_style,
        style=table_line_color,
    )

    table.add_column()
    if dashboard_table:
        table.add_column(max_width=25, no_wrap=True)
    else:
        table.add_column(min_width=60, overflow="fold")

    if replica_thread_id:
        table.add_row("[#c5c7d2]Host", "%s" % dolphie.replica_connections[replica_thread_id]["host"])
    else:
        table.add_row("[#c5c7d2]Primary", "%s" % data["Master_Host"])

    if not dashboard_table:
        table.add_row("[#c5c7d2]User", "%s" % data["Master_User"])

    table.add_row(
        "[#c5c7d2]Thread",
        "[#c5c7d2]IO %s [#c5c7d2]SQL %s" % (data["Slave_IO_Running"], data["Slave_SQL_Running"]),
    )
    if data["Seconds_Behind_Master"] is None:
        table.add_row("[#c5c7d2]Lag", "")
    else:
        lag_source = "Lag"
        if sbm_source:
            lag_source = f"Lag ({sbm_source})"

        table.add_row(
            "[#c5c7d2]%s" % lag_source,
            "%s [#c5c7d2]Speed[/#c5c7d2] %s" % (lag, speed),
        )

    replication_status_filtering = [
        "Replicate_Do_DB",
        "Replicate_Ignore_Table",
        "Replicate_Do_Table",
        "Replicate_Wild_Do_Table",
        "Replicate_Wild_Ignore_Table",
    ]

    if not replica_thread_id:
        for status_filter in replication_status_filtering:
            value = dolphie.replication_status.get(status_filter)
            if value:
                table.add_row(f"[#c5c7d2]{filter}", str(value))

    if dashboard_table:
        table.add_row("[#c5c7d2]Binlog IO", "%s" % (data["Master_Log_File"]))
        table.add_row("[#c5c7d2]Binlog SQL", "%s" % (data["Relay_Master_Log_File"]))
        table.add_row("[#c5c7d2]Relay Log ", "%s" % (data["Relay_Log_File"]))
        table.add_row("[#c5c7d2]GTID", "%s" % gtid_status)
        table.add_row("[#c5c7d2]State", "%s" % data["Slave_SQL_Running_State"])
    else:
        table.add_row(
            "[#c5c7d2]Binlog IO",
            "%s ([#969aad]%s[/#969aad])" % (data["Master_Log_File"], data["Read_Master_Log_Pos"]),
        )
        table.add_row(
            "[#c5c7d2]Binlog SQL",
            "%s ([#969aad]%s[/#969aad])"
            % (
                data["Relay_Master_Log_File"],
                data["Exec_Master_Log_Pos"],
            ),
        )
        table.add_row(
            "[#c5c7d2]Relay Log",
            "%s ([#969aad]%s[/#969aad])" % (data["Relay_Log_File"], data["Relay_Log_Pos"]),
        )

        if mysql_gtid_enabled:
            primary_uuid = data["Master_UUID"]
            executed_gtid_set = data["Executed_Gtid_Set"]
            retrieved_gtid_set = data["Retrieved_Gtid_Set"]

            table.add_row("[#c5c7d2]Auto Position", "%s" % data["Auto_Position"])

            # We find errant transactions for replicas here
            if replica_thread_id:
                # We ignore the GTID set that relates to the replica's primary UUID
                def remove_primary_uuid_gtid_set(gtid_set):
                    lines = gtid_set.splitlines()
                    filtered_lines = [line for line in lines if dolphie.server_uuid not in line]

                    # If the last line ends with a comma, remove comma since that query would fail
                    last_line = filtered_lines[-1]
                    if filtered_lines and last_line.endswith(","):
                        last_line = last_line[:-1]

                    return "\n".join(filtered_lines)

                replica_gtid_set = remove_primary_uuid_gtid_set(executed_gtid_set)
                primary_gtid_set = remove_primary_uuid_gtid_set(dolphie.global_variables["gtid_executed"])

                db_cursor.execute(f"SELECT GTID_SUBTRACT('{replica_gtid_set}', '{primary_gtid_set}') AS errant_trxs")
                gtid_data = db_cursor.fetchone()
                if gtid_data:
                    if gtid_data["errant_trxs"]:
                        errant_trx = f"[#fc7979]{gtid_data['errant_trxs']}[/#fc7979]"
                    else:
                        errant_trx = "[#54efae]None[/#54efae]"

                    table.add_row("[#c5c7d2]Errant TRX", "%s" % errant_trx)

                # Since this is for the replica view, we use the host's UUID since its the primary
                primary_uuid = dolphie.server_uuid

            def color_gtid_set(match):
                source_id = match.group(1)
                transaction_id = match.group(2)

                if source_id == primary_uuid:
                    return f"[#91abec]{source_id}[/#91abec]:{transaction_id}"
                else:
                    return f"[#969aad]{source_id}:{transaction_id}[/#969aad]"

            # Example GTID: 3beacd96-6fe3-18ec-9d95-b4592zec4b45:1-26
            pattern = re.compile(r"\b(\w+(?:-\w+){4}):(\d+(?:-\d*)?)\b")
            retrieved_gtid_set = pattern.sub(color_gtid_set, retrieved_gtid_set.replace(",", ""))
            executed_gtid_set = pattern.sub(color_gtid_set, executed_gtid_set.replace(",", ""))

            table.add_row("[#c5c7d2]Retrieved GTID", "%s" % retrieved_gtid_set)
            table.add_row("[#c5c7d2]Executed GTID", "%s" % executed_gtid_set)
        elif mariadb_gtid_enabled:
            table.add_row("[#c5c7d2]GTID IO Pos", "%s" % data["Gtid_IO_Pos"])

        error_types = ["Last_Error", "Last_IO_Error", "Last_SQL_Error"]
        errors = [(error_type, data[error_type]) for error_type in error_types if data[error_type]]

        if errors:
            for error_type, error_message in errors:
                table.add_row("[#c5c7d2]%s" % error_type.replace("_", " "), "%s" % error_message)
        else:
            table.add_row("[#c5c7d2]IO State", "%s" % data["Slave_IO_State"])
            table.add_row("[#c5c7d2]SQL State", "%s" % data["Slave_SQL_Running_State"])

    return table


def fetch_replica_table_data(dolphie: Dolphie):
    replica_tables = {}
    for row in dolphie.replica_data:
        thread_id = row["id"]

        # Resolve IPs to addresses and add to cache for fast lookup
        host = dolphie.get_hostname(row["host"].split(":")[0])

        try:
            if thread_id not in dolphie.replica_connections:
                dolphie.replica_connections[thread_id] = {
                    "host": host,
                    "connection": pymysql.connect(
                        host=host,
                        user=dolphie.user,
                        passwd=dolphie.password,
                        port=dolphie.port,
                        ssl=dolphie.ssl,
                        autocommit=True,
                    ),
                    "cursor": None,
                    "previous_sbm": 0,
                }

            replica_connection = dolphie.replica_connections[thread_id]
            replica_connection["cursor"] = replica_connection["connection"].cursor(pymysql.cursors.DictCursor)
            replica_connection["cursor"].execute(MySQLQueries.replication_status)
            replica_data = replica_connection["cursor"].fetchone()

            if replica_data:
                replica_tables[host] = create_table(dolphie, data=replica_data, replica_thread_id=thread_id)
        except pymysql.Error as e:
            table = Table(box=box.ROUNDED, show_header=False, style="#b0bad7")

            table.add_column()
            table.add_column()

            table.add_row("[#c5c7d2]Host", host)
            table.add_row("[#c5c7d2]User", row["user"])
            table.add_row("[#fc7979]Error", e.args[1])

            replica_tables[host] = table

    return replica_tables



#!/usr/bin/env python3

# ****************************
# *        Dolphie           *
# * Author: Charles Thompson *
# ****************************

import os
import re
import sys
from argparse import ArgumentParser, RawTextHelpFormatter
from configparser import ConfigParser
from datetime import datetime
from urllib.parse import urlparse

import dolphie.Modules.MetricManager as MetricManager
import myloginpath
from dolphie import Dolphie
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Panels import dashboard_panel, processlist_panel, replication_panel
from dolphie.Widgets.topbar import TopBar
from rich.console import Console
from rich.prompt import Prompt
from rich.traceback import Traceback
from textual import events, on, work
from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, VerticalScroll
from textual.css.query import NoMatches
from textual.widgets import (
    DataTable,
    Label,
    LoadingIndicator,
    Sparkline,
    Static,
    Switch,
    TabbedContent,
    TabPane,
)
from textual.worker import Worker, WorkerState


def parse_args(dolphie: Dolphie):
    epilog = """
Config file with [client] section supports these options:
    host
    user
    password
    port
    socket
    ssl_mode REQUIRED/VERIFY_CA/VERIFY_IDENTITY
    ssl_ca
    ssl_cert
    ssl_key

Login path file supports these options:
    host
    user
    password
    port
    socket

Environment variables support these options:
    DOLPHIE_USER
    DOLPHIE_PASSWORD
    DOLPHIE_HOST
    DOLPHIE_PORT
    DOLPHIE_SOCKET

"""
    parser = ArgumentParser(
        conflict_handler="resolve",
        description="Dolphie, an intuitive feature-rich top tool for monitoring MySQL in real time",
        epilog=epilog,
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "uri",
        metavar="uri",
        type=str,
        nargs="?",
        help=(
            "Use a URI string for credentials - format: mysql://user:password@host:port (port is optional with"
            " default 3306)"
        ),
    )

    parser.add_argument(
        "-u",
        "--user",
        dest="user",
        type=str,
        help="Username for MySQL",
    )
    parser.add_argument("-p", "--password", dest="password", type=str, help="Password for MySQL")
    parser.add_argument(
        "--ask-pass",
        dest="ask_password",
        action="store_true",
        default=False,
        help="Ask for password (hidden text)",
    )
    parser.add_argument(
        "-h",
        "--host",
        dest="host",
        type=str,
        help="Hostname/IP address for MySQL",
    )
    parser.add_argument(
        "-P",
        "--port",
        dest="port",
        type=int,
        help="Port for MySQL (Socket has precendence)",
    )
    parser.add_argument(
        "-S",
        "--socket",
        dest="socket",
        type=str,
        help="Socket file for MySQL",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        type=str,
        help=(
            "Config file path to use. This should use [client] section. "
            "See below for options support [default: ~/.my.cnf]"
        ),
    )
    parser.add_argument(
        "-f",
        "--host-cache-file",
        dest="host_cache_file",
        type=str,
        help=(
            "Resolve IPs to hostnames when your DNS is unable to. Each IP/hostname pair should be on its own line "
            "using format: ip=hostname [default: ~/dolphie_host_cache]"
        ),
    )
    parser.add_argument(
        "-q",
        "--quick-switch-hosts-file",
        dest="quick_switch_hosts_file",
        type=str,
        help=(
            "Specify where the file is that stores the hosts you connect to for quick switching [default:"
            " ~/dolphie_quick_switch_hosts]"
        ),
    )
    parser.add_argument(
        "-l",
        "--login-path",
        dest="login_path",
        default="client",
        type=str,
        help=(
            "Specify login path to use mysql_config_editor's file ~/.mylogin.cnf for encrypted login credentials. "
            "Supercedes config file [default: %(default)s]"
        ),
    )
    parser.add_argument(
        "-r",
        "--refresh_interval",
        dest="refresh_interval",
        default=1,
        type=int,
        help="How much time to wait in seconds between each refresh [default: %(default)s]",
    )
    parser.add_argument(
        "-H",
        "--heartbeat-table",
        dest="heartbeat_table",
        type=str,
        help=(
            "If your hosts use pt-heartbeat, specify table in format db.table to use the timestamp it "
            "has for replication lag instead of Seconds_Behind_Master from SHOW SLAVE STATUS"
        ),
    )
    parser.add_argument(
        "--ssl-mode",
        dest="ssl_mode",
        type=str,
        help=(
            "Desired security state of the connection to the host. Supports: "
            "REQUIRED/VERIFY_CA/VERIFY_IDENTITY [default: OFF]"
        ),
    )
    parser.add_argument(
        "--ssl-ca",
        dest="ssl_ca",
        type=str,
        help="Path to the file that contains a PEM-formatted CA certificate",
    )
    parser.add_argument(
        "--ssl-cert",
        dest="ssl_cert",
        type=str,
        help="Path to the file that contains a PEM-formatted client certificate",
    )
    parser.add_argument(
        "--ssl-key",
        dest="ssl_key",
        type=str,
        help="Path to the file that contains a PEM-formatted private key for the client certificate",
    )
    parser.add_argument(
        "--hide-dashboard",
        dest="hide_dashboard",
        action="store_true",
        default=False,
        help=(
            "Start without showing dashboard. This is good to use if you want to reclaim terminal space and "
            "not execute the additional queries for it"
        ),
    )
    parser.add_argument(
        "--show-trxs-only",
        dest="show_trxs_only",
        action="store_true",
        default=False,
        help="Start with only showing threads that have an active transaction",
    )
    parser.add_argument(
        "--additional-columns",
        dest="show_additional_query_columns",
        action="store_true",
        default=False,
        help="Start with additional columns in Processlist panel",
    )
    parser.add_argument(
        "--use-processlist",
        dest="use_processlist",
        action="store_true",
        default=False,
        help="Start with using Processlist instead of Performance Schema for listing queries",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=dolphie.app_version, help="Display version and exit"
    )

    console = Console(style="indian_red", highlight=False)

    home_dir = os.path.expanduser("~")

    parameter_options = vars(parser.parse_args())  # Convert object to dict
    basic_options = ["user", "password", "host", "port", "socket"]

    dolphie.config_file = f"{home_dir}/.my.cnf"
    if parameter_options["config_file"]:
        dolphie.config_file = parameter_options["config_file"]

    # Use config file for login credentials
    if os.path.isfile(dolphie.config_file):
        cfg = ConfigParser()
        cfg.read(dolphie.config_file)

        for option in basic_options:
            if cfg.has_option("client", option):
                setattr(dolphie, option, cfg.get("client", option))

        if cfg.has_option("client", "ssl_mode"):
            ssl_mode = cfg.get("client", "ssl_mode").upper()

            if ssl_mode == "REQUIRED":
                dolphie.ssl[""] = True
            elif ssl_mode == "VERIFY_CA":
                dolphie.ssl["check_hostname"] = False
            elif ssl_mode == "VERIFY_IDENTITY":
                dolphie.ssl["check_hostname"] = True
            else:
                sys.exit(console.print(f"Unsupported SSL mode [b]{ssl_mode}[/b]"))

        if cfg.has_option("client", "ssl_ca"):
            dolphie.ssl["ca"] = cfg.get("client", "ssl_ca")
        if cfg.has_option("client", "ssl_cert"):
            dolphie.ssl["cert"] = cfg.get("client", "ssl_cert")
        if cfg.has_option("client", "ssl_key"):
            dolphie.ssl["key"] = cfg.get("client", "ssl_key")

    # Use login path for login credentials
    if parameter_options["login_path"]:
        try:
            login_path_data = myloginpath.parse(parameter_options["login_path"])

            for option in basic_options:
                if option in login_path_data:
                    setattr(dolphie, option, login_path_data[option])
        except Exception as e:
            # Don't error out for default login path
            if parameter_options["login_path"] != "client":
                sys.exit(console.print(f"Problem reading login path file: {e}"))

    # Use environment variables for basic options if specified
    for option in basic_options:
        environment_var = "DOLPHIE_%s" % option.upper()
        if environment_var in os.environ and os.environ[environment_var]:
            setattr(dolphie, option, os.environ[environment_var])

    # Use parameter options if specified
    for option in basic_options:
        if parameter_options[option]:
            setattr(dolphie, option, parameter_options[option])

    # Lastly, parse URI if specified
    if parameter_options["uri"]:
        try:
            parsed = urlparse(parameter_options["uri"])

            if parsed.scheme != "mysql":
                sys.exit(
                    console.print("Invalid URI scheme: Only 'mysql' is supported (see --help for more information)")
                )

            dolphie.user = parsed.username
            dolphie.password = parsed.password
            dolphie.host = parsed.hostname
            dolphie.port = parsed.port or 3306
        except Exception as e:
            sys.exit(console.print(f"Invalid URI: {e} (see --help for more information)"))

    if parameter_options["ask_password"]:
        dolphie.password = Prompt.ask("[b #bbc8e8]Password", password=True)

    if not dolphie.host:
        dolphie.host = "localhost"

    if parameter_options["refresh_interval"]:
        dolphie.refresh_interval = parameter_options["refresh_interval"]

    if parameter_options["heartbeat_table"]:
        pattern_match = re.search(r"^(\w+\.\w+)$", parameter_options["heartbeat_table"])
        if pattern_match:
            dolphie.heartbeat_table = parameter_options["heartbeat_table"]
            MySQLQueries.heartbeat_replica_lag = MySQLQueries.heartbeat_replica_lag.replace(
                "$placeholder", dolphie.heartbeat_table
            )
        else:
            sys.exit(console.print("Your heartbeat table did not conform to the proper format: db.table"))

    if parameter_options["ssl_mode"]:
        ssl_mode = parameter_options["ssl_mode"].upper()

        if ssl_mode == "REQUIRED":
            dolphie.ssl[""] = True
        elif ssl_mode == "VERIFY_CA":
            dolphie.ssl["check_hostame"] = False
        elif ssl_mode == "VERIFY_IDENTITY":
            dolphie.ssl["check_hostame"] = True
        else:
            sys.exit(console.print(f"Unsupported SSL mode [b]{ssl_mode}[/b]"))

    if parameter_options["ssl_ca"]:
        dolphie.ssl["ca"] = parameter_options["ssl_ca"]
    if parameter_options["ssl_cert"]:
        dolphie.ssl["cert"] = parameter_options["ssl_cert"]
    if parameter_options["ssl_key"]:
        dolphie.ssl["key"] = parameter_options["ssl_key"]

    dolphie.host_cache_file = f"{home_dir}/dolphie_host_cache"
    if parameter_options["host_cache_file"]:
        dolphie.host_cache_file = parameter_options["host_cache_file"]

    dolphie.quick_switch_hosts_file = f"{home_dir}/dolphie_quick_switch_hosts"
    if parameter_options["quick_switch_hosts_file"]:
        dolphie.quick_switch_hosts_file = parameter_options["quick_switch_hosts_file"]

    dolphie.show_trxs_only = parameter_options["show_trxs_only"]
    dolphie.show_additional_query_columns = parameter_options["show_additional_query_columns"]
    dolphie.use_processlist = parameter_options["use_processlist"]
    dolphie.hide_dashboard = parameter_options["hide_dashboard"]

    if os.path.exists(dolphie.quick_switch_hosts_file):
        with open(dolphie.quick_switch_hosts_file, "r") as file:
            dolphie.quick_switch_hosts = [line.strip() for line in file]


class DolphieApp(App):
    TITLE = "Dolphie"
    CSS_PATH = "Dolphie.css"

    def __init__(self, dolphie: Dolphie):
        super().__init__()

        self.dolphie = dolphie
        dolphie.app = self

        self.console.set_window_title(self.TITLE)

    @work(exclusive=True, thread=True)
    def worker_fetch_data(self):
        dolphie = self.dolphie

        if dolphie.quick_switched_connection:
            self.quick_host_switch()

        if dolphie.metric_manager.worker_start_time:
            dolphie.metric_manager.update_metrics_with_last_value()

        try:
            if not dolphie.main_db_connection or not dolphie.main_db_connection.connection.open:
                dolphie.db_connect()

            dolphie.worker_start_time = datetime.now()
            dolphie.worker_job_time = (dolphie.worker_start_time - dolphie.worker_previous_start_time).total_seconds()
            dolphie.worker_previous_start_time = dolphie.worker_start_time

            dolphie.global_variables = dolphie.main_db_connection.fetch_data("variables")
            dolphie.global_status = dolphie.main_db_connection.fetch_data("status")
            dolphie.innodb_metrics = dolphie.main_db_connection.fetch_data("innodb_metrics")
            dolphie.replica_data = dolphie.main_db_connection.fetch_data(
                "find_replicas", dolphie.use_performance_schema
            )
            dolphie.fetch_replication_data()
            dolphie.massage_metrics_data()

            if dolphie.display_dashboard_panel:
                dolphie.binlog_status = dolphie.main_db_connection.fetch_data("binlog_status")

            if dolphie.display_replication_panel:
                dolphie.replica_tables = replication_panel.fetch_replica_table_data(dolphie)

            if dolphie.display_processlist_panel:
                dolphie.processlist_threads = processlist_panel.fetch_data(dolphie)

            # If we're not displaying the replication panel, close all replica connections
            if not dolphie.display_replication_panel and dolphie.replica_connections:
                for connection in dolphie.replica_connections.values():
                    connection["connection"].close()

                dolphie.replica_connections = {}
                dolphie.replica_data = {}
                dolphie.replica_tables = {}

            dolphie.metric_manager.refresh_data(
                worker_start_time=dolphie.worker_start_time,
                worker_job_time=dolphie.worker_job_time,
                global_variables=dolphie.global_variables,
                global_status=dolphie.global_status,
                innodb_metrics=dolphie.innodb_metrics,
                replication_status=dolphie.replication_status,
                replication_lag=dolphie.replica_lag,
            )
        except ManualException as e:
            self.exit(message=e.output())

    def on_worker_state_changed(self, event: Worker.StateChanged):
        if event.state == WorkerState.SUCCESS:
            dolphie = self.dolphie

            # Skip this if the conditions are right
            if (
                len(self.screen_stack) > 1
                or dolphie.pause_refresh
                or not self.dolphie.main_db_connection.connection.open
                or dolphie.quick_switched_connection
            ):
                self.set_timer(0.5, self.worker_fetch_data)
                return

            try:
                loading_indicator = self.app.query_one("LoadingIndicator")
                if loading_indicator.display:
                    loading_indicator.display = False
                    self.app.query_one("#main_container").display = True
                    self.layout_graphs()

                    # Update our header with host information
                    self.app.query_one("#topbar_host").update(f"{dolphie.mysql_host}:{dolphie.port}")

                if dolphie.display_dashboard_panel:
                    self.refresh_panel("dashboard")

                    # Update the sparkline for queries per second
                    sparkline = self.app.query_one("#panel_dashboard_queries_qps")
                    sparkline_data = dolphie.metric_manager.metrics.dml.Queries.values
                    if not sparkline.display:
                        sparkline_data = [0]
                        sparkline.display = True

                    sparkline.data = sparkline_data
                    sparkline.refresh()

                if dolphie.display_processlist_panel:
                    self.refresh_panel("processlist")

                if dolphie.display_replication_panel:
                    self.refresh_panel("replication")

                if dolphie.display_graphs_panel:
                    # Add/remove replication tab based on replication status
                    replication_tab = self.app.query_one("#tabbed_content", TabbedContent)
                    if dolphie.replication_status:
                        if not self.app.query("#tab_replication_lag"):
                            replication_tab.add_pane(
                                pane=TabPane(
                                    "Replication",
                                    Label(id="stats_replication_lag", classes="stats_data"),
                                    MetricManager.Graph(id="graph_replication_lag", classes="panel_data"),
                                    id="tab_replication_lag",
                                )
                            )
                    else:
                        if self.app.query("#tab_replication_lag"):
                            replication_tab.remove_pane("tab_replication_lag")

                    # Refresh the graph(s) for the selected tab
                    active_graph = self.query_one("#tabbed_content", TabbedContent).active
                    metric_instance_name = active_graph.split("tab_")[1]
                    self.update_graphs(metric_instance_name)

                # We take a snapshot of the processlist to be used for commands
                # since the data can change after a key is pressed
                dolphie.processlist_threads_snapshot = dolphie.processlist_threads.copy()
            except NoMatches:
                # This is thrown if a user toggles panels on and off and the display_* states aren't 1:1
                # with worker thread/state change due to asynchronous nature of the worker thread
                pass

            self.set_timer(self.dolphie.refresh_interval, self.worker_fetch_data)

    def on_key(self, event: events.Key):
        if len(self.screen_stack) > 1:
            return

        self.dolphie.capture_key(event.key)

    def on_mount(self):
        dolphie = self.dolphie

        dolphie.load_host_cache_file()

        # Set these components by default to not show
        components_to_disable = [
            ".panel_container",
            "#panel_processlist",
            "Sparkline",
            "#footer",
        ]
        exempt_components = {"panel_dashboard", "panel_processlist"}

        for component in components_to_disable:
            display_off_components = self.query(component)
            for display_off_component in display_off_components:
                if display_off_component.id not in exempt_components:
                    display_off_component.display = False

        dolphie.display_processlist_panel = True
        if dolphie.hide_dashboard:
            self.query_one("#panel_dashboard", Container).display = False
            dolphie.display_dashboard_panel = False
        else:
            dolphie.display_dashboard_panel = True

        # Set default switches to be toggled on
        switches = self.query(".switch_container Switch")
        switches_to_toggle = [switch for switch in switches if switch.id not in ["Queries", "Threads_connected"]]
        for switch in switches_to_toggle:
            switch.toggle()

        dolphie.check_for_update()

        self.worker_fetch_data()

    def _handle_exception(self, error: Exception) -> None:
        self.bell()

        # We have a ManualException class that we use to output errors in a nice format
        if error.__class__.__name__ == "ManualException":
            message = error.output()
        else:
            message = Traceback(show_locals=True, width=None, locals_max_length=5)

        self.exit(message=message)

    @on(TabbedContent.TabActivated)
    def tab_changed(self, event: TabbedContent.TabActivated):
        if len(self.screen_stack) > 1:
            return

        metric_instance_name = event.tab.id.split("tab_")[1]
        self.update_graphs(metric_instance_name)

    @on(Switch.Changed)
    def switch_changed(self, event: Switch.Changed):
        if len(self.screen_stack) > 1:
            return

        metric_instance_name = event.switch.name
        metric = event.switch.id

        metric_instance = getattr(self.dolphie.metric_manager.metrics, metric_instance_name)
        metric_data: MetricManager.MetricData = getattr(metric_instance, metric)
        metric_data.visible = event.value

        self.update_graphs(metric_instance_name)

    def generate_switches(self, metric_instance_name):
        metric_instance = getattr(self.dolphie.metric_manager.metrics, metric_instance_name)

        for metric, metric_data in metric_instance.__dict__.items():
            if isinstance(metric_data, MetricManager.MetricData) and metric_data.graphable:
                yield Label(metric_data.label)
                yield Switch(animate=False, id=metric, name=metric_instance_name)

    def update_graphs(self, tab_metric_instance_name):
        for metric_instance in self.dolphie.metric_manager.metrics.__dict__.values():
            if tab_metric_instance_name == metric_instance.tab_name:
                for graph_name in metric_instance.graphs:
                    self.query_one(f"#{graph_name}").render_graph(metric_instance)

        self.update_stats_label(tab_metric_instance_name)

    def update_stats_label(self, tab_metric_instance_name):
        stat_data = {}

        for metric_instance in self.dolphie.metric_manager.metrics.__dict__.values():
            if hasattr(metric_instance, "tab_name") and metric_instance.tab_name == tab_metric_instance_name:
                number_format_func = MetricManager.get_number_format_function(metric_instance, color=True)
                for metric_data in metric_instance.__dict__.values():
                    if isinstance(metric_data, MetricManager.MetricData) and metric_data.values and metric_data.visible:
                        stat_data[metric_data.label] = number_format_func(metric_data.values[-1])

        formatted_stat_data = "  ".join(f"[b #bbc8e8]{label}[/b #bbc8e8] {value}" for label, value in stat_data.items())
        self.query_one(f"#stats_{tab_metric_instance_name}").update(formatted_stat_data)

    def refresh_panel(self, panel_name, toggled=False):
        # If loading indicator is displaying, don't refresh
        if self.app.query_one("LoadingIndicator").display:
            return

        if panel_name == "replication":
            # When replication panel status is changed, we need to refresh the dashboard panel as well since
            # it adds/removes it from there
            self.query_one("#panel_replication_data", Static).update(replication_panel.create_panel(self.dolphie))

            if toggled and self.dolphie.replication_status:
                self.query_one("#panel_dashboard_data", Static).update(dashboard_panel.create_panel(self.dolphie))

        elif panel_name == "dashboard":
            self.query_one("#panel_dashboard_data", Static).update(dashboard_panel.create_panel(self.dolphie))
        elif panel_name == "processlist":
            processlist_panel.create_panel(self.dolphie)

    def quick_host_switch(self):
        dolphie = self.dolphie

        dolphie.main_db_connection.connection.close()
        dolphie.secondary_db_connection.connection.close()

        dolphie.replication_status = {}
        dolphie.replica_data = {}
        dolphie.replica_tables = {}

        if dolphie.replica_connections:
            for connection in dolphie.replica_connections.values():
                connection["connection"].close()

            dolphie.replica_connections = {}

        dolphie.metric_manager.reset()
        dolphie.dolphie_start_time = datetime.now()

        # Set the graph switches to what they're currently selected to since we reset metric_manager
        switches = self.query(".switch_container Switch")
        for switch in switches:
            switch: Switch
            metric_instance_name = switch.name
            metric = switch.id

            metric_instance = getattr(self.dolphie.metric_manager.metrics, metric_instance_name)
            metric_data: MetricManager.MetricData = getattr(metric_instance, metric)
            metric_data.visible = switch.value

        dolphie.quick_switched_connection = False

    def layout_graphs(self):
        if self.dolphie.is_mysql_version_at_least("8.0.30"):
            self.query_one("#graph_redo_log").styles.width = "55%"
            self.query_one("#graph_redo_log_bar").styles.width = "12%"
            self.query_one("#graph_redo_log_active_count").styles.width = "33%"
            self.dolphie.metric_manager.metrics.redo_log_active_count.Active_redo_log_count.visible = True
            self.query_one("#graph_redo_log_active_count").display = True
        else:
            self.query_one("#graph_redo_log").styles.width = "88%"
            self.query_one("#graph_redo_log_bar").styles.width = "12%"
            self.query_one("#graph_redo_log_active_count").display = False

        self.query_one("#graph_adaptive_hash_index").styles.width = "50%"
        self.query_one("#graph_adaptive_hash_index_hit_ratio").styles.width = "50%"

    def compose(self) -> ComposeResult:
        yield TopBar(host="Connecting to MySQL", app_version=self.dolphie.app_version, help="press [b]?[/b] for help")

        yield LoadingIndicator()
        with VerticalScroll(id="main_container"):
            with Container(id="panel_dashboard", classes="panel_container"):
                yield Static(id="panel_dashboard_data", classes="panel_data")
                yield Sparkline([], id="panel_dashboard_queries_qps")

            with Container(id="panel_graphs", classes="panel_container"):
                with TabbedContent(initial="tab_dml", id="tabbed_content"):
                    with TabPane("DML", id="tab_dml"):
                        yield Label(id="stats_dml", classes="stats_data")
                        yield MetricManager.Graph(id="graph_dml", classes="panel_data")

                        with Horizontal(classes="switch_container"):
                            yield from self.generate_switches("dml")

                    with TabPane("Table Cache", id="tab_table_cache"):
                        yield Label(id="stats_table_cache", classes="stats_data")
                        yield MetricManager.Graph(id="graph_table_cache", classes="panel_data")

                        with Horizontal(classes="switch_container"):
                            yield from self.generate_switches("table_cache")

                    with TabPane("Threads", id="tab_threads"):
                        yield Label(id="stats_threads", classes="stats_data")
                        yield MetricManager.Graph(id="graph_threads", classes="panel_data")

                        with Horizontal(classes="switch_container"):
                            yield from self.generate_switches("threads")

                    with TabPane("Buffer Pool Requests", id="tab_buffer_pool_requests"):
                        yield Label(id="stats_buffer_pool_requests", classes="stats_data")
                        yield MetricManager.Graph(id="graph_buffer_pool_requests", classes="panel_data")

                        with Horizontal(classes="switch_container"):
                            yield from self.generate_switches("buffer_pool_requests")
                    with TabPane("Checkpoint", id="tab_checkpoint"):
                        yield Label(id="stats_checkpoint", classes="stats_data")
                        yield MetricManager.Graph(id="graph_checkpoint", classes="panel_data")

                    with TabPane("Redo Log", id="tab_redo_log"):
                        yield Label(id="stats_redo_log", classes="stats_data")
                        with Horizontal():
                            yield MetricManager.Graph(id="graph_redo_log", classes="panel_data")
                            yield MetricManager.Graph(id="graph_redo_log_active_count", classes="panel_data")
                            yield MetricManager.Graph(bar=True, id="graph_redo_log_bar", classes="panel_data")

                    with TabPane("Adaptive Hash Index", id="tab_adaptive_hash_index"):
                        yield Label(id="stats_adaptive_hash_index", classes="stats_data")
                        with Horizontal():
                            yield MetricManager.Graph(id="graph_adaptive_hash_index", classes="panel_data")
                            yield MetricManager.Graph(id="graph_adaptive_hash_index_hit_ratio", classes="panel_data")

                        with Horizontal(classes="switch_container"):
                            yield from self.generate_switches("adaptive_hash_index")

            with VerticalScroll(id="panel_replication", classes="panel_container"):
                yield Static(id="panel_replication_data", classes="panel_data")

            yield DataTable(id="panel_processlist", classes="panel_data", show_cursor=False)

            yield Static(id="footer")


def main():
    # Set environment variables so Textual can use all the pretty colors
    os.environ["TERM"] = "xterm-256color"
    os.environ["COLORTERM"] = "truecolor"

    dolphie = Dolphie()
    parse_args(dolphie)

    app = DolphieApp(dolphie)
    app.run()


if __name__ == "__main__":
    main()



from textual.app import ComposeResult
from textual.containers import Container
from textual.reactive import reactive
from textual.widgets import Label


class TopBar(Container):
    app_version = reactive("", init=False)
    host = reactive("", init=False)

    def __init__(self, app_version="", host="", help=""):
        super().__init__()

        self.help = "press any key to return (except navigation keys)"
        if help:
            self.help = help

        self.topbar_title = Label(
            f" [b #bbc8e8]dolphie[/b #bbc8e8] :dolphin: [#bbc8e8]v{app_version}", id="topbar_title"
        )
        self.topbar_host = Label(self.host, id="topbar_host")
        self.topbar_help = Label(self.help, id="topbar_help")

        self.app_version = app_version
        self.host = host

    def watch_app_version(self):
        self.topbar_title.update(f" [b #bbc8e8]dolphie[/b #bbc8e8] :dolphin: [#bbc8e8]v{self.app_version}")

    def watch_host(self):
        self.topbar_host.update(self.host)

    def compose(self) -> ComposeResult:
        yield self.topbar_title
        yield self.topbar_host
        yield self.topbar_help



import pymysql
from dolphie.Modules.Functions import detect_encoding
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.Queries import MySQLQueries


class Database:
    def __init__(self, host, user, password, socket, port, ssl):
        self.host = host
        self.user = user
        self.password = password
        self.socket = socket
        self.port = port
        self.ssl = ssl

        try:
            self.connection = pymysql.connect(
                host=host,
                user=user,
                passwd=password,
                unix_socket=socket,
                port=port,
                use_unicode=False,
                ssl=ssl,
                autocommit=True,
            )
            self.cursor = self.connection.cursor(pymysql.cursors.DictCursor)
        except pymysql.Error as e:
            raise ManualException(f"Failed to connect to database host {self.host}:{self.port}", reason=e.args[1])
        except FileNotFoundError:  # Catch SSL file path errors
            raise ManualException("SSL certificate file path isn't valid!")

    def execute(self, query, values=None, ignore_error=False):
        if not self.connection.open:
            return

        # Prefix all queries with dolphie so they can be identified in the processlist from other people
        query = "/* dolphie */ " + query

        try:
            return self.cursor.execute(query, values)
        except Exception as e:
            if ignore_error:
                return None
            else:
                raise ManualException("Failed to execute query\n", query=query, reason=e.args[1])

    def process_row(self, row):
        processed_row = {}

        for field, value in row.items():
            if isinstance(value, (bytes, bytearray)):
                if "query" in field:
                    processed_row[field] = value.decode(detect_encoding(value))
                else:
                    processed_row[field] = value.decode()
            else:
                processed_row[field] = value

        return processed_row

    def fetchall(self):
        rows = [self.process_row(row) for row in self.cursor.fetchall()]

        if not rows:
            return []

        return rows

    def fetchone(self):
        row = self.cursor.fetchone()

        if not row:
            return []

        return self.process_row(row)

    def fetch_value_from_field(self, query, field=None, values=None):
        self.execute(query, values)
        data = self.cursor.fetchone()

        if not data:
            return None

        field = field or next(iter(data))  # Use field if provided, otherwise get first field
        value = data.get(field)

        if isinstance(value, (bytes, bytearray)):
            if field == "Status":
                return value.decode(detect_encoding(value))
            return value.decode()
        return value

    def fetch_data(self, command, performance_schema=None):
        command_data = {}

        if command in {"status", "variables"}:
            self.execute(getattr(MySQLQueries, command))
            data = self.fetchall()

            for row in data:
                variable = row["Variable_name"]
                value = row["Value"]

                converted_value = int(value) if value.isnumeric() else value

                command_data[variable] = converted_value

        elif command == "find_replicas":
            query = MySQLQueries.ps_find_replicas if performance_schema else MySQLQueries.pl_find_replicas

            self.execute(query)
            command_data = self.fetchall()

        elif command == "innodb_metrics":
            self.execute(MySQLQueries.innodb_metrics)
            data = self.fetchall()

            for row in data:
                metric = row["NAME"]
                value = int(row["COUNT"])

                command_data[metric] = value

        else:
            self.execute(getattr(MySQLQueries, command))
            command_data = self.fetchone()

        return command_data



from dolphie.Widgets.topbar import TopBar
from textual import events
from textual.app import ComposeResult
from textual.containers import VerticalScroll
from textual.screen import Screen
from textual.widgets import Static


class CommandScreen(Screen):
    CSS = """
        CommandScreen VerticalScroll {
            padding: 1;
        }
    """

    def __init__(self, app_version, host, data):
        super().__init__()
        self.app_version = app_version
        self.host = host
        self.data = data

    def on_key(self, event: events.Key):
        exclude_events = ["up", "down", "left", "right", "pageup", "pagedown", "home", "end", "tab", "enter"]
        if event.key not in exclude_events:
            if self.screen.is_attached:
                self.app.pop_screen()

    def compose(self) -> ComposeResult:
        yield TopBar(app_version=self.app_version, host=self.host)
        yield VerticalScroll(Static(self.data))



import re
from datetime import timedelta

from dolphie import Dolphie
from dolphie.Modules.Functions import format_number, format_time
from dolphie.Modules.Queries import MySQLQueries
from rich.text import Text
from textual.widgets import DataTable


def create_panel(dolphie: Dolphie) -> DataTable:
    columns = [
        {"name": "Thread ID", "field": "id", "width": 11, "format_number": False},
        {"name": "Username", "field": "user", "width": 13, "format_number": False},
    ]

    if dolphie.show_additional_query_columns:
        columns.extend(
            [
                {"name": "Hostname/IP", "field": "host", "width": 16, "format_number": False},
                {"name": "Database", "field": "db", "width": 13, "format_number": False},
            ]
        )

    columns.extend(
        [
            {"name": "Command", "field": "command", "width": 8, "format_number": False},
            {"name": "State", "field": "state", "width": 16, "format_number": False},
            {"name": "TRX State", "field": "trx_state", "width": 9, "format_number": False},
            {"name": "Rows Lock", "field": "trx_rows_locked", "width": 9, "format_number": True},
            {"name": "Rows Mod", "field": "trx_rows_modified", "width": 8, "format_number": True},
        ]
    )

    if (
        dolphie.show_additional_query_columns
        and "innodb_thread_concurrency" in dolphie.global_variables
        and dolphie.global_variables["innodb_thread_concurrency"]
    ):
        columns.append({"name": "Tickets", "field": "trx_concurrency_tickets", "width": 8, "format_number": True})

    columns.extend(
        [
            {"name": "Time", "field": "formatted_time", "width": 9, "format_number": False},
            {"name": "Query", "field": "query", "width": None, "format_number": False},
        ]
    )

    processlist_datatable = dolphie.app.query_one("#panel_processlist", DataTable)

    # Clear table if columns change
    if len(processlist_datatable.columns) != len(columns):
        processlist_datatable.clear(columns=True)

    # Add columns to the datatable if it is empty
    if not processlist_datatable.columns:
        for column_data in columns:
            column_name = column_data["name"]
            column_key = column_data["field"]
            column_width = column_data["width"]
            processlist_datatable.add_column(column_name, key=column_key, width=column_width)

    # Iterate through dolphie.processlist_threads
    for thread_id, thread in dolphie.processlist_threads.items():
        # Add or modify the "command" field based on the condition
        if thread["command"] == "Killed":
            thread["command"] = "[#fc7979]Killed"

        # Check if the thread_id exists in the datatable
        if thread_id in processlist_datatable.rows:
            datatable_row = processlist_datatable.get_row(thread_id)

            # Update the datatable if values differ
            for column_id, column_data in enumerate(columns):
                column_name = column_data["field"]
                column_format_number = column_data["format_number"]

                update_width = False
                if column_name == "query":
                    value = re.sub(r"\s+", " ", thread[column_name])
                    update_width = True
                else:
                    value = format_number(thread[column_name]) if column_format_number else thread[column_name]

                if value != datatable_row[column_id] or column_name == "formatted_time":
                    processlist_datatable.update_cell(thread_id, column_name, value, update_width=update_width)
        else:
            # Add a new row to the datatable if thread_id does not exist
            row_values = []
            for column_data in columns:
                column_name = column_data["field"]
                column_format_number = column_data["format_number"]

                if column_name == "query":
                    value = re.sub(r"\s+", " ", thread[column_name])
                else:
                    value = format_number(thread[column_name]) if column_format_number else thread[column_name]

                row_values.append(value)

            processlist_datatable.add_row(*row_values, key=thread_id)

    # Remove rows from processlist_datatable that no longer exist in dolphie.processlist_threads
    rows_to_remove = set(processlist_datatable.rows.keys()) - set(dolphie.processlist_threads.keys())
    for id in rows_to_remove:
        processlist_datatable.remove_row(id)

    processlist_datatable.sort("formatted_time", reverse=dolphie.sort_by_time_descending)


def fetch_data(dolphie: Dolphie):
    if dolphie.use_performance_schema:
        processlist_query = MySQLQueries.ps_query
    else:
        processlist_query = MySQLQueries.pl_query

    ########################
    # WHERE clause filters #
    ########################
    where_clause = []

    # Filter out idle threads if specified
    if not dolphie.show_idle_threads:
        if dolphie.use_performance_schema:
            where_clause.append(
                "(processlist_command != 'Sleep' AND processlist_command NOT LIKE 'Binlog Dump%') AND "
                "(processlist_info IS NOT NULL OR trx_query IS NOT NULL)"
            )
        else:
            where_clause.append(
                "(Command != 'Sleep' AND Command NOT LIKE 'Binlog Dump%') AND "
                "(Info IS NOT NULL OR trx_query IS NOT NULL)"
            )

    # Only show running transactions only
    if dolphie.show_trxs_only:
        where_clause.append("trx_state != ''")

    # Filter user
    if dolphie.user_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_user = '%s'" % dolphie.user_filter)
        else:
            where_clause.append("User = '%s'" % dolphie.user_filter)

    # Filter database
    if dolphie.db_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_db = '%s'" % dolphie.db_filter)
        else:
            where_clause.append("db = '%s'" % dolphie.db_filter)

    # Filter hostname/IP
    if dolphie.host_filter:
        # Have to use LIKE since there's a port at the end
        if dolphie.use_performance_schema:
            where_clause.append("processlist_host LIKE '%s%%'" % dolphie.host_filter)
        else:
            where_clause.append("Host LIKE '%s%%'" % dolphie.host_filter)

    # Filter time
    if dolphie.query_time_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_time >= '%s'" % dolphie.query_time_filter)
        else:
            where_clause.append("Time >= '%s'" % dolphie.query_time_filter)

    # Filter query
    if dolphie.query_filter:
        if dolphie.use_performance_schema:
            where_clause.append(
                "(processlist_info LIKE '%%%s%%' OR trx_query LIKE '%%%s%%')"
                % (dolphie.query_filter, dolphie.query_filter),
            )
        else:
            where_clause.append("Info LIKE '%%%s%%'" % dolphie.query_filter)

    # Add in our dynamic WHERE clause for filtering
    if where_clause:
        processlist_query = processlist_query.replace("$placeholder", "AND " + " AND ".join(where_clause))
    else:
        processlist_query = processlist_query.replace("$placeholder", "")

    processlist_threads = {}
    # Run the processlist query
    dolphie.main_db_connection.execute(processlist_query)
    threads = dolphie.main_db_connection.fetchall()

    for thread in threads:
        # Don't include Dolphie's threads
        if dolphie.main_db_connection_id == thread["id"] or dolphie.secondary_db_connection_id == thread["id"]:
            continue

        command = thread["command"]
        # Use trx_query over Performance Schema query since it's more accurate
        if dolphie.use_performance_schema and thread["trx_query"]:
            query = thread["trx_query"]
        else:
            query = thread["query"]

        # Determine time color
        time = int(thread["time"])
        thread_color = ""
        if "SELECT /*!40001 SQL_NO_CACHE */ *" in query:
            thread_color = "magenta"
        elif query:
            if time >= 10:
                thread_color = "#fc7979"
            elif time >= 5:
                thread_color = "#f1fb82"
            else:
                thread_color = "#54efae"

        formatted_time = TextPlus(format_time(time), style=thread_color)
        formatted_time_with_days = TextPlus("{:0>8}".format(str(timedelta(seconds=time))), style=thread_color)

        host = thread["host"].split(":")[0]
        host = dolphie.get_hostname(host)

        mysql_thread_id = thread.get("mysql_thread_id")

        processlist_threads[str(thread["id"])] = {
            "id": str(thread["id"]),
            "mysql_thread_id": mysql_thread_id,
            "user": thread["user"],
            "host": host,
            "db": thread["db"],
            "time": time,
            "formatted_time_with_days": formatted_time_with_days,
            "formatted_time": formatted_time,
            "command": command,
            "state": thread["state"],
            "trx_state": thread["trx_state"],
            "trx_operation_state": thread["trx_operation_state"],
            "trx_rows_locked": thread["trx_rows_locked"],
            "trx_rows_modified": thread["trx_rows_modified"],
            "trx_concurrency_tickets": thread["trx_concurrency_tickets"],
            "query": query,
        }

    return processlist_threads


class TextPlus(Text):
    """Custom patch for a Rich `Text` object to allow Textual `DataTable`
    sorting when a Text object is included in a row."""

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Text):
            return NotImplemented
        return self.plain < other.plain

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Text):
            return NotImplemented
        return self.plain <= other.plain

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Text):
            return NotImplemented
        return self.plain > other.plain

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Text):
            return NotImplemented
        return self.plain >= other.plain



from datetime import datetime, timedelta

from dolphie import Dolphie
from dolphie.Modules.Functions import format_bytes, format_number
from dolphie.Panels import replication_panel
from rich import box
from rich.style import Style
from rich.table import Table


def create_panel(dolphie: Dolphie) -> Table:
    global_status = dolphie.global_status
    global_variables = dolphie.global_variables
    binlog_status = dolphie.binlog_status

    tables_to_add = []
    uptime = str(timedelta(seconds=global_status["Uptime"]))

    dashboard_grid = Table.grid()
    dashboard_grid.add_column()
    dashboard_grid.add_column()
    dashboard_grid.add_column()

    table_title_style = Style(bold=True)
    table_box = box.ROUNDED
    table_line_color = "#52608d"

    ################
    # Information #
    ###############
    table_information = Table(
        show_header=False,
        box=table_box,
        title="Host Information",
        title_style=table_title_style,
        style=table_line_color,
    )

    if dolphie.worker_job_time < 1:
        refresh_latency = 0
    else:
        refresh_latency = round(dolphie.worker_job_time - dolphie.refresh_interval, 2)

    if global_variables["read_only"] == "ON":
        if not dolphie.replication_status:
            global_variables["read_only"] = "YES ([indian_red]SHOULD BE NO?[/indian_red])"
        else:
            global_variables["read_only"] = "YES"
    elif global_variables["read_only"] == "OFF":
        global_variables["read_only"] = "NO"

    runtime = str(datetime.now() - dolphie.dolphie_start_time).split(".")[0]

    table_information.add_column()
    table_information.add_column(width=25)
    table_information.add_row("[#c5c7d2]Version", f"{dolphie.host_distro} {dolphie.mysql_version}")
    table_information.add_row(
        "[#c5c7d2]", "%s (%s)" % (global_variables["version_compile_os"], global_variables["version_compile_machine"])
    )
    table_information.add_row("[#c5c7d2]Uptime", uptime)
    table_information.add_row("[#c5c7d2]Runtime", f"{runtime} [#c5c7d2]latency:[/#c5c7d2] {refresh_latency}s")
    table_information.add_row("[#c5c7d2]Read Only", global_variables["read_only"])
    table_information.add_row("[#c5c7d2]Replicas", "%s" % len(dolphie.replica_data))
    table_information.add_row(
        "[#c5c7d2]Threads",
        "[#c5c7d2]con[/#c5c7d2] %s[#91abec]/[/#91abec][#c5c7d2]run[/#c5c7d2]"
        " %s[#91abec]/[/#91abec][#c5c7d2]cac[/#c5c7d2] %s"
        % (
            format_number(global_status["Threads_connected"]),
            format_number(global_status["Threads_running"]),
            format_number(global_status["Threads_cached"]),
        ),
    )
    table_information.add_row(
        "[#c5c7d2]Tables",
        "[#c5c7d2]open[/#c5c7d2] %s[#91abec]/[/#91abec][#c5c7d2]opened[/#c5c7d2] %s"
        % (
            format_number(global_status["Open_tables"]),
            format_number(global_status["Opened_tables"]),
        ),
    )

    tables_to_add.append(table_information)

    ###########
    # InnoDB  #
    ###########
    table_innodb = Table(
        show_header=False,
        box=table_box,
        title="InnoDB",
        title_style=table_title_style,
        style=table_line_color,
    )

    table_innodb.add_column()
    table_innodb.add_column(width=9)

    history_list_length = dolphie.innodb_metrics.get("trx_rseg_history_len", "N/A")

    # Calculate InnoDB memory read hit efficiency
    ib_pool_disk_reads = global_status.get("Innodb_buffer_pool_reads", 0)
    ib_pool_mem_reads = global_status.get(
        "Innodb_buffer_pool_read_requests", 1
    )  # Default to 1 to avoid division by zero

    if ib_pool_disk_reads >= ib_pool_mem_reads:
        innodb_efficiency = "[#fc7979]0.00%"
    else:
        efficiency = 100 - (ib_pool_disk_reads / ib_pool_mem_reads * 100)

        if efficiency > 90:
            color_code = "#54efae"
        elif efficiency > 80:
            color_code = "#f1fb82"
        else:
            color_code = "#fc7979"

        innodb_efficiency = f"[{color_code}]{efficiency:.2f}%"

    # Add data to our table
    table_innodb.add_row("[#c5c7d2]Read Hit", innodb_efficiency)
    table_innodb.add_row("[#c5c7d2]Chkpt Age", dolphie.metric_manager.get_metric_checkpoint_age(format=True))
    table_innodb.add_row("[#c5c7d2]AHI Hit", dolphie.metric_manager.get_metric_adaptive_hash_index())

    bp_instances = global_variables.get("innodb_buffer_pool_instances", "N/A")
    if bp_instances != "N/A":
        plural = "s" if bp_instances > 1 else ""
        table_innodb.add_row("[#c5c7d2]BP Instance" + plural, format_number(bp_instances))
    else:
        table_innodb.add_row("[#c5c7d2]BP Instance", bp_instances)

    table_innodb.add_row("[#c5c7d2]BP Size", format_bytes(global_variables["innodb_buffer_pool_size"]))
    table_innodb.add_row(
        "[#c5c7d2]BP Available",
        format_bytes(
            dolphie.global_variables["innodb_buffer_pool_size"] - dolphie.global_status["Innodb_buffer_pool_bytes_data"]
        ),
    )
    table_innodb.add_row("[#c5c7d2]BP Dirty", format_bytes(global_status["Innodb_buffer_pool_bytes_dirty"]))
    table_innodb.add_row("[#c5c7d2]History List", format_number(history_list_length))

    tables_to_add.append(table_innodb)

    ##############
    # Binary Log #
    ##############
    table_primary = Table()

    if binlog_status:
        table_primary = Table(
            show_header=False,
            box=table_box,
            title="Binary Log",
            title_style=table_title_style,
            style=table_line_color,
        )

        if dolphie.previous_binlog_position == 0:
            diff_binlog_position = 0
        elif dolphie.previous_binlog_position > binlog_status["Position"]:
            diff_binlog_position = "Binlog Rotated"
        else:
            diff_binlog_position = format_bytes(binlog_status["Position"] - dolphie.previous_binlog_position)

        binlog_cache = 100
        binlog_cache_disk = global_status["Binlog_cache_disk_use"]
        binlog_cache_mem = global_status["Binlog_cache_use"]
        if binlog_cache_disk and binlog_cache_mem:
            if binlog_cache_disk >= binlog_cache_mem:
                innodb_efficiency = "[#fc7979]0.00%"
            elif binlog_cache_mem > binlog_cache_disk:
                binlog_cache = round(100 - (binlog_cache_disk / binlog_cache_mem), 2)

        table_primary.add_column()
        table_primary.add_column(max_width=40)
        table_primary.add_row("[#c5c7d2]File name", binlog_status["File"])
        table_primary.add_row(
            "[#c5c7d2]Position",
            "%s" % (str(binlog_status["Position"])),
        )
        table_primary.add_row(
            "[#c5c7d2]Size",
            "%s" % format_bytes(binlog_status["Position"]),
        )
        table_primary.add_row("[#c5c7d2]Diff", str(diff_binlog_position))
        table_primary.add_row("[#c5c7d2]Cache Hit", f"{binlog_cache}%")

        binlog_format = global_variables.get("binlog_format", "N/A")
        if binlog_format == "ROW":
            binlog_row_image = global_variables.get("binlog_row_image", "N/A")
            table_primary.add_row("[#c5c7d2]Format", "{} ({})".format(binlog_format, binlog_row_image))
        else:
            table_primary.add_row("[#c5c7d2]Format", binlog_format, binlog_row_image)

        gtid_mode = global_variables.get("gtid_mode", "N/A")
        table_primary.add_row("[#c5c7d2]GTID", gtid_mode)

        binlog_compression = global_variables.get("binlog_transaction_compression", "N/A")
        table_primary.add_row("[#c5c7d2]Compression", binlog_compression)

        tables_to_add.append(table_primary)

        # Save some global_variables to be used in next refresh
        if dolphie.binlog_status:
            dolphie.previous_binlog_position = dolphie.binlog_status["Position"]

    ###############
    # Replication #
    ###############
    if dolphie.replication_status and not dolphie.display_replication_panel:
        tables_to_add.append(replication_panel.create_table(dolphie, dashboard_table=True))

    ###############
    # Statisitics #
    ###############
    table_stats = Table(
        show_header=False,
        box=table_box,
        title="Statisitics/s",
        title_style=table_title_style,
        style=table_line_color,
    )

    table_stats.add_column()
    table_stats.add_column(min_width=7)

    table_stats.add_row("[#c5c7d2]Queries", dolphie.metric_manager.get_metric_calculate_per_sec("Queries"))
    table_stats.add_row("[#c5c7d2]SELECT", dolphie.metric_manager.get_metric_calculate_per_sec("Com_select"))
    table_stats.add_row("[#c5c7d2]INSERT", dolphie.metric_manager.get_metric_calculate_per_sec("Com_insert"))
    table_stats.add_row("[#c5c7d2]UPDATE", dolphie.metric_manager.get_metric_calculate_per_sec("Com_update"))
    table_stats.add_row("[#c5c7d2]DELETE", dolphie.metric_manager.get_metric_calculate_per_sec("Com_delete"))
    table_stats.add_row("[#c5c7d2]REPLACE", dolphie.metric_manager.get_metric_calculate_per_sec("Com_replace"))
    table_stats.add_row("[#c5c7d2]COMMIT", dolphie.metric_manager.get_metric_calculate_per_sec("Com_commit"))
    table_stats.add_row("[#c5c7d2]ROLLBACK", dolphie.metric_manager.get_metric_calculate_per_sec("Com_rollback"))

    tables_to_add.append(table_stats)

    dashboard_grid.add_row(*tables_to_add)

    return dashboard_grid



from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Checkbox, Input, Label, RadioButton, RadioSet
from textual_autocomplete import AutoComplete, Dropdown, DropdownItem


class CommandModal(ModalScreen):
    CSS = """
        CommandModal > Vertical {
            background: #121626;
            border: thick #20263d;
            height: auto;
            width: auto;
        }
        CommandModal > Vertical > * {
            width: auto;
            height: auto;
        }
        CommandModal #kill_container {
            width: 100%;
            height: 6;
        }
        CommandModal Label {
            text-style: bold;
            width: 100%;
            content-align: center middle;
            padding-bottom: 1;
        }
    """
    BINDINGS = [
        Binding("escape", "app.pop_screen", "", show=False),
    ]

    def __init__(
        self, message, variable=None, processlist_data=None, show_filter_options=False, show_kill_options=False
    ):
        super().__init__()
        self.variable = variable
        self.message = message
        self.show_filter_options = show_filter_options
        self.show_kill_options = show_kill_options
        self.processlist_data = processlist_data

        self.dropdown_items = []
        if processlist_data and (not show_filter_options and not show_kill_options):
            sorted_keys = sorted(processlist_data.keys(), key=lambda x: int(x))
            self.dropdown_items = [DropdownItem(id) for id in sorted_keys]

    def compose(self) -> ComposeResult:
        with Vertical():
            with Vertical():
                yield Label(self.message)
                with RadioSet(id="filter_radio_buttons"):
                    yield RadioButton("User", id="user")
                    yield RadioButton("Host/IP", id="host")
                    yield RadioButton("Database", id="database")
                    yield RadioButton("Query Text", id="query_text")
                    yield RadioButton("Query Time", id="query_time")
                with Vertical(id="kill_container"):
                    with RadioSet(id="kill_radio_buttons"):
                        yield RadioButton("Username", id="username")
                        yield RadioButton("Host/IP", id="host")
                        yield RadioButton("Time range", id="time_range")
                    yield Checkbox("Include sleeping queries", id="sleeping_queries")
                yield AutoComplete(
                    Input(id="modal_input"),
                    Dropdown(id="dropdown_items", items=self.dropdown_items),
                )
            with Horizontal(classes="button_container"):
                yield Button("Submit", id="submit", variant="primary")
                yield Button("Cancel", id="cancel")

    def on_mount(self):
        input = self.query_one("#modal_input", Input)
        filter_radio_buttons = self.query_one("#filter_radio_buttons", RadioSet)
        kill_container = self.query_one("#kill_container", Vertical)

        filter_radio_buttons.display = False
        kill_container.display = False

        if self.show_filter_options:
            filter_radio_buttons.display = True
            filter_radio_buttons.focus()

            input.placeholder = "Select an option from above"
        elif self.show_kill_options:
            sleeping_queries_checkbox = self.query_one("#sleeping_queries", Checkbox)
            sleeping_queries_checkbox.toggle()

            kill_container.display = True

            kill_radio_buttons = self.query_one("#kill_radio_buttons", RadioSet)
            kill_radio_buttons.focus()

            input.placeholder = "Select an option from above"
        else:
            input.focus()

    def on_input_submitted(self):
        self.query_one("#submit", Button).press()

    def create_dropdown_items(self, field):
        self.dropdown_items = []

        if field:
            sorted_array = sorted(set(data.get(field) for _, data in self.processlist_data.items()))
            self.dropdown_items = [DropdownItem(value) for value in sorted_array]

        self.query_one("#dropdown_items", Dropdown).items = self.dropdown_items

    def on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        modal_input = self.query_one("#modal_input", Input)

        self.create_dropdown_items(None)  # empty string to clear dropdown items

        if self.show_filter_options:
            if event.pressed.id == "database":
                self.create_dropdown_items("db")
                modal_input.placeholder = "Database name"
            elif event.pressed.id == "host":
                self.create_dropdown_items("host")
                modal_input.placeholder = "Hostname or IP address"
            elif event.pressed.id == "query_text":
                modal_input.placeholder = "Partial query text"
            elif event.pressed.id == "query_time":
                modal_input.placeholder = "Query time (in seconds)"
            elif event.pressed.id == "user":
                self.create_dropdown_items("user")
                modal_input.placeholder = "Username"
        elif self.show_kill_options:
            if event.pressed.id == "username":
                self.create_dropdown_items("user")
                modal_input.placeholder = "Username"
            elif event.pressed.id == "host":
                self.create_dropdown_items("host")
                modal_input.placeholder = "Hostname or IP address"
            elif event.pressed.id == "time_range":
                modal_input.placeholder = "Time range (ex. 10-20)"

        modal_input.focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "submit":
            modal_input = self.query_one("#modal_input", Input)

            if self.variable:
                self.dismiss([self.variable, modal_input.value])
            elif self.show_filter_options:
                for rb in self.query("#filter_radio_buttons RadioButton"):
                    if rb.value:
                        self.dismiss([rb.id, modal_input.value])
            elif self.show_kill_options:
                checkbox_sleeping_queries = self.query_one("#sleeping_queries", Checkbox)
                for rb in self.query("#kill_container RadioButton"):
                    if rb.value:
                        self.dismiss([rb.id, modal_input.value, checkbox_sleeping_queries.value])
            else:
                self.dismiss(modal_input.value)
        else:
            self.app.pop_screen()
